import collections
import functools
import warnings

# Make sure we're showing DeprecationWarnings
warnings.filterwarnings('default','',DeprecationWarning)

class deprecated(object):
    '''Decorator factory class which returns a decorator function that
    marks a function as deprecated. It will result in a warning being
    emitted when the function is used, once per invocation point and
    deprecated function. The doc string of the deprecated function is
    overwritten with a deprecation message.

    @deprecated(use_instead="c_to_p(...)')
    def hgvsc_to_hgvsp():

    results in warnings like this:

    /home/reece/projects/hgvs/hgvs/variantmapper.py:280: DeprecationWarning: Call to deprecated function hgvsc_to_hgvsp; use c_to_p(...) instead

    '''

    def __init__(self,use_instead=None):
        # only warn once per caller location and function
        self.seen = collections.Counter()
        self.use_instead = use_instead

    def __call__(self,func):
        msg = "Call to deprecated function {}".format(func.__name__)
        if self.use_instead:
            msg += '; use '+self.use_instead+' instead'

        def wrapper(*args, **kwargs):
            fingerprint = (func.__name__,func.func_code.co_filename,func.func_code.co_firstlineno)
            if fingerprint not in self.seen:
                warnings.warn_explicit(
                    msg,
                    category=DeprecationWarning,
                    filename=func.func_code.co_filename,
                    lineno=func.func_code.co_firstlineno + 1)
            self.seen.update([fingerprint])
            return func(*args, **kwargs)
        wrapper.__doc__ = "Deprecated"
        if self.use_instead:
            wrapper.__doc__ += '; use '+self.use_instead+' instead'
        return wrapper



#
# Utility class for creating an hgvsp SequenceVariant object,
# given a transcript with variants applied.
# Used in hgvsc to hgvsp conversion.
#
import collections
import difflib

import hgvs.edit
import hgvs.exceptions
import hgvs.location
import hgvs.posedit
import hgvs.utils
import hgvs.variant

DBG = False

class AltSeqToHgvsp(object):

    def __init__(self, ref_data, alt_data):
        """Constructor

        :param ref_data: reference transcript record
        :type ref_data: recordtype
        :param alt_data: alt transcript record
        :type ref_data: recordtype
        """
        self._ref_data = ref_data
        self._alt_data = alt_data
        self._protein_accession = self._ref_data.protein_accession
        self._ref_seq = self._ref_data.aa_sequence
        self._alt_seq = self._alt_data.aa_sequence
        self._is_frameshift = self._alt_data.is_frameshift
        self._frameshift_start = self._alt_data.frameshift_start
        self._is_substitution = self._alt_data.is_substitution
        self._is_ambiguous = self._alt_data.is_ambiguous

        if DBG:
            print "len ref seq:{} len alt seq:{}".format(len(self._ref_seq), len(self._alt_seq))
            print "fs start:{} protein ac:{}".format(self._frameshift_start, self._protein_accession)
            print self._ref_seq
            print self._alt_seq
            print "aa variant start: {}".format(self._alt_data.variant_start_aa)
            print self._ref_data.transcript_sequence
            print self._alt_data.transcript_sequence

    def build_hgvsp(self):
        """Compare two amino acid sequences; generate an hgvs tag from the output

        :return list of variants in sequence order
        :rtype list of dict
        """

        variants = []

        if not self._is_ambiguous and len(self._alt_seq) > 0:

            do_delins = True
            if self._ref_seq == self._alt_seq:
                do_delins = False
            elif self._is_substitution:
                if len(self._ref_seq) == len(self._alt_seq):
                    diff_pos = [(i, self._ref_seq[i], self._alt_seq[i]) for i in xrange(len(self._ref_seq))
                                if  self._ref_seq[i] != self._alt_seq[i]]
                    if len(diff_pos) == 1:
                        (start, deletion, insertion) = diff_pos[0]
                        variants.append({"start": start + 1, "ins": insertion, "del": deletion})
                        do_delins = False

                elif self._alt_seq[self._alt_data.variant_start_aa - 1] == "*" and \
                                self._ref_seq[self._alt_data.variant_start_aa - 1] != "*": # introduced stop codon
                    deletion = self._ref_seq[self._alt_data.variant_start_aa - 1:]
                    variants.append({"start": self._alt_data.variant_start_aa, "ins": "*", "del": deletion})
                    do_delins = False

            if do_delins:
                if self._alt_data.is_frameshift:
                    start = self._alt_data.variant_start_aa - 1
                    aa_start = self._alt_data.variant_start_aa
                    while self._ref_seq[start] == self._alt_seq[start]:
                        start += 1
                        aa_start += 1
                    insertion = list(self._alt_seq[start:])
                    deletion = list(self._ref_seq[start:])
                    variants.append({"start": aa_start, "ins": insertion, "del": deletion})

                else: # non-frameshifting delins or dup
                    # get size diff from diff in ref/alt lengths
                    start = self._alt_data.variant_start_aa - 1
                    aa_start = self._alt_data.variant_start_aa
                    delta = len(self._alt_seq) - len(self._ref_seq)
                    while self._ref_seq[start] == self._alt_seq[start]:
                        start += 1
                        aa_start += 1
                    offset = start + abs(delta)

                    if delta > 0:   # net insertion
                        insertion = list(self._alt_seq[start:offset])
                        deletion = []
                        ref_sub = self._ref_seq[start:]
                        alt_sub = self._alt_seq[offset:]
                    elif delta < 0: # net deletion
                        insertion = []
                        deletion = list(self._ref_seq[start:offset])
                        ref_sub = self._ref_seq[offset:]
                        alt_sub = self._alt_seq[start:]
                    else:
                        insertion = []
                        deletion = []
                        ref_sub = self._ref_seq[start:]
                        alt_sub = self._alt_seq[start:]

                    # from start, get del/ins out to last difference
                    diff_indices = [i for i in xrange(len(ref_sub)) if ref_sub[i] != alt_sub[i]]
                    if diff_indices:
                        max_diff = diff_indices[-1] + 1
                        insertion.extend(list(alt_sub[:max_diff]))
                        deletion.extend(list(ref_sub[:max_diff]))

                    variants.append({"start": aa_start, "ins": insertion, "del": deletion})

            if DBG:
                print variants

        if self._is_ambiguous:
            var_ps = [self._create_variant('', '', '', '', acc=self._protein_accession,
                                           is_ambiguous=self._is_ambiguous)]
        elif len(self._alt_seq) == 0:
            var_ps = [self._create_variant('', '', '', '', acc=self._protein_accession,
                                           is_ambiguous=self._is_ambiguous, is_no_protein=True)]
        elif variants:
            var_ps = [self._convert_to_sequence_variants(x, self._protein_accession) for x in variants]
        else:    # ref = alt - "silent" hgvs change
            var_ps = [self._create_variant('', '', '', '', acc=self._protein_accession)]

        # TODO - handle multiple variants

        if len(var_ps) > 1:
            raise hgvs.exceptions.HGVSError("Got multiple AA variants - not supported")
        return var_ps[0]

    #
    # internal methods
    #

    def _convert_to_sequence_variants(self, variant, acc):
        """Convert AA variant to an hgvs representation

        :param variant: contains start, del, and ins
        :type variant: dict
        :param acc: protein accession
        :type acc: str
        :return hgvs string
        :rtype str
        """
        start = variant['start']
        insertion = ''.join(variant['ins'])
        deletion = ''.join(variant['del'])

        # defaults
        is_dup = False  # assume not dup
        fsext_len = None    # fs or ext length
        is_sub = False
        is_ext = False

        if start == 1:                                                  # initial methionine is modified
            aa_start = aa_end = hgvs.location.AAPosition(base=start, aa=deletion)
            ref = ''
            alt = ''
            self._is_ambiguous = True   # side-effect

        if insertion and insertion.find("*") == 0:                      # stop codon at variant position
            aa_start = aa_end = hgvs.location.AAPosition(base=start, aa=deletion[0])
            ref = ''
            alt = '*'
            is_sub = True

        elif start == len(self._ref_seq):                               # extension
            if self._alt_seq[-1] == '*':
                fsext_len = len(insertion) - len(deletion) # don't include the former stop codon
            else:
                fsext_len = '?'
            subst_at_stop_codon = insertion[0]

            aa_start = aa_end = hgvs.location.AAPosition(base=start, aa='*')
            ref = ''
            alt = subst_at_stop_codon
            is_ext = True

        elif self._is_frameshift:                                       # frameshift
            aa_start = aa_end = hgvs.location.AAPosition(base=start, aa=deletion[0])
            ref = ''

            try:
                fsext_len = str(insertion.index("*") + 1)    # start w/ 1st change; ends w/ * (inclusive)
            except ValueError:
                fsext_len = "?"

            alt = insertion[0]

        else:                                                           # no frameshift - sub/delins/dup
            if len(insertion) == len(deletion) == 1:                    # substitution
                aa_start = aa_end = hgvs.location.AAPosition(base=start, aa=deletion)
                ref = ''
                alt = insertion
                is_sub = True

            elif len(deletion) > 0:                                     # delins OR deletion OR stop codon at variant position
                ref = deletion
                end = start + len(deletion) - 1
                if len(insertion) > 0:                                  # delins
                    aa_start = hgvs.location.AAPosition(base=start, aa=deletion[0])
                    if end > start:
                        aa_end =  hgvs.location.AAPosition(base=end, aa=deletion[-1])
                    else:
                        aa_end = aa_start
                    alt = insertion

                else:                                                   # deletion OR stop codon at variant position
                    if len(deletion) + start == len(self._ref_seq):     # stop codon at variant position
                        aa_start = hgvs.location.AAPosition(base=start, aa=deletion[0])
                        aa_end = hgvs.location.AAPosition(base=start, aa=deletion[0])
                        ref = ''
                        alt = '*'
                        is_sub = True
                    else:                                               # deletion
                        aa_start = hgvs.location.AAPosition(base=start, aa=deletion[0])
                        if end > start:
                            aa_end =  hgvs.location.AAPosition(base=end, aa=deletion[-1])
                        else:
                            aa_end = aa_start
                        alt = None

            elif len(deletion) == 0:                                    # insertion OR duplication OR extension

                is_dup, dup_start = self._check_if_ins_is_dup(start, insertion)

                if is_dup:                                              # duplication
                    dup_end = dup_start + len(insertion) - 1
                    aa_start = hgvs.location.AAPosition(base=dup_start, aa=insertion[0])
                    aa_end =  hgvs.location.AAPosition(base=dup_end, aa=insertion[-1])
                    ref = alt = None

                else:                                                   # insertion
                    start -= 1
                    end = start + 1

                    aa_start = hgvs.location.AAPosition(base=start, aa=self._ref_seq[start - 1])
                    aa_end =  hgvs.location.AAPosition(base=end, aa=self._ref_seq[end - 1])
                    ref = None
                    alt = insertion

            else: # should never get here
                raise ValueError("unexpected variant: {}".format(variant))

        var_p = self._create_variant(aa_start, aa_end, ref, alt,
                                     fsext_len=fsext_len, is_dup=is_dup, acc=acc, is_ambiguous=self._is_ambiguous,
                                     is_sub=is_sub, is_ext=is_ext)

        return var_p

    def _check_if_ins_is_dup(self, start, insertion):
        """Helper to identify an insertion as a duplicate

        :param start: 1-based insertion start
        :type start: int
        :param insertion: sequence
        :type insertion: str
        :return (is duplicate, variant start)
        :rtype (bool, int)
        """
        is_dup = False  # assume no
        variant_start = None

        dup_candidate_start = start - len(insertion) - 1
        dup_candidate = self._ref_seq[dup_candidate_start: dup_candidate_start + len(insertion)]
        if insertion == dup_candidate:
            is_dup = True
            variant_start = dup_candidate_start + 1

        return is_dup, variant_start

    def _create_variant(self, start, end, ref, alt, fsext_len=None, is_dup=False, acc=None,
                        is_ambiguous=False, is_sub=False, is_ext=False, is_no_protein=False):
        """Creates a SequenceVariant object"""
        interval = hgvs.location.Interval(start=start, end=end)
        # Note - order matters
        if is_no_protein:
            edit = '0'
        elif is_ambiguous:
            edit = '?'
        elif is_sub:
            edit = hgvs.edit.AASub(ref=ref, alt=alt)
        elif is_ext:
            edit = hgvs.edit.AAExt(ref=ref, alt=alt, aaterm='*', length=fsext_len)
        elif self._is_frameshift:
            edit = hgvs.edit.AAFs(ref=ref, alt=alt, length=fsext_len)
        elif is_dup:
            edit = hgvs.edit.Dup()
        elif ref == alt == '':
            edit = '='
        else:
            edit = hgvs.edit.AARefAlt(ref=ref, alt=alt)
        posedit = hgvs.posedit.PosEdit(interval, edit)
        if not(is_ambiguous and start == ''):
            posedit._set_uncertain()
        var_p = hgvs.variant.SequenceVariant(acc, 'p', posedit)

        return var_p


## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# tests running from a local sqlite instance

import csv
import os
import unittest

import hgvs.dataproviders.uta

import hgvs.variantmapper as variantmapper
import hgvs.parser


def gcp_file_reader(fn):
    rdr = csv.DictReader(open(fn, 'r'), delimiter='\t')
    for rec in rdr:
        if rec['id'].startswith('#'):
            continue
        yield rec


class TestVariantMapperFast(unittest.TestCase):
    def setUp(self):
        sqlite_path = 'sqlite://{}'.format(os.path.join(os.path.dirname(__file__), 'db', 'uta-test-1.db'))
        self.hdp = hgvs.dataproviders.uta.connect(sqlite_path)
        self.hm = hgvs.variantmapper.VariantMapper(self.hdp, cache_transcripts=True)
        self.hp = hgvs.parser.Parser()
        self.failed = []

    def test_fast(self):
        for rec in gcp_file_reader('tests/data/sqlite_test_gcp.tsv'):
            self._test_gcp_mapping(rec)

    def _test_gcp_mapping(self, rec):
        var_g = self.hp.parse_hgvs_variant(rec['HGVSg'])
        var_c = self.hp.parse_hgvs_variant(rec['HGVSc'])
        var_p = self.hp.parse_hgvs_variant(rec['HGVSp']) if rec['HGVSp'] is not None and rec['HGVSp'] != '' else None

        # g -> c
        var_c_test = self.hm.g_to_c(var_g, var_c.ac)
        self.assertEquals(str(var_c_test), str(var_c))

        # c -> g
        var_g_test = self.hm.c_to_g(var_c, var_g.ac)
        self.assertEquals(str(var_g_test), str(var_g))

        if var_p is not None:
            # c -> p
            var_p_test = self.hm.c_to_p(var_c, var_p.ac)
            hgvs_p_test = str(var_p_test)
            self.assertEquals(hgvs_p_test, str(var_p))


if __name__ == '__main__':
    unittest.main()

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



import logging
import os
import sqlite3
import urlparse

#TODO: make dynamic import with importlib
import psycopg2
import psycopg2.extras

from .interface import Interface
from ..utils.aminoacids import seq_md5


localhost_db_url = 'postgresql://localhost/uta'
public_db_url = 'postgresql://uta_public:uta_public@uta.invitae.com/uta'
default_db_url = os.environ.get('UTA_DB_URL',public_db_url)
pg_schema = 'uta_20140210'

logger = logging.getLogger(__name__)


def connect(db_url=default_db_url):
    """
    Connect to a UTA database instance and return a UTA interface instance.

    When called with an explicit db_url argument, that db_url is used for connecting.

    When called without an explicit argument, the function default is
    determined by the environment variable UTA_DB_URL if it exists, or
    hgvs.datainterface.uta.public_db_url otherwise.

    The format of the db_url is driver://user:pass@host/database (the same
    as that used by SQLAlchemy).  Examples:

    A remote public postgresql database:
        postgresql://uta_public:uta_public@uta.invitae.com/uta'

    A local postgresql database:
        postgresql://localhost/uta
    
    A local SQLite database:
      sqlite:////tmp/uta-0.0.6.db
    """

    url = urlparse.urlparse(db_url)
    
    if url.scheme == 'sqlite':
        conn = UTA_sqlite(url)
    elif url.scheme == 'postgresql':
        conn = UTA_postgresql(url)
    else:
        # fell through connection scheme cases
        raise RuntimeError("{url.scheme} in {db_url} is not supported".format(url=url, db_url=db_url))

    conn.db_url = db_url
    return conn


class UTABase(Interface):
    sql = {
        "acs_for_protein_md5": """
            select ac
            from seq_anno
            where seq_id=?
            """,

        "gene_info": """
            select *
            from gene
            where hgnc=?
            """,

        "tx_exons": """
            select *
            from tx_exon_aln_v
            where tx_ac=? and alt_ac=? and alt_aln_method=?
            order by alt_start_i
            """,
        
        "tx_for_gene": """
            select hgnc, cds_start_i, cds_end_i, tx_ac, alt_ac, alt_aln_method
            from transcript T
            join exon_set ES on T.ac=ES.tx_ac where alt_aln_method != 'transcript' and hgnc=?
            """,

        "tx_for_region": """
            select tx_ac,alt_ac,alt_strand,alt_aln_method,min(start_i) as start_i,max(end_i) as end_i
            from exon_set ES
            join exon E on ES.exon_set_id=E.exon_set_id 
            where alt_ac=? and alt_aln_method=?
            group by tx_ac,alt_ac,alt_strand,alt_aln_method
            having max(end_i)>=? and min(start_i)<=?
            """,

        "tx_identity_info": """
            select distinct(tx_ac), alt_ac, alt_aln_method, cds_start_i, cds_end_i, lengths, hgnc
            from tx_def_summary_v
            where tx_ac=?
            """,

        "tx_info": """
            select hgnc, cds_start_i, cds_end_i, tx_ac, alt_ac, alt_aln_method
            from transcript T
            join exon_set ES on T.ac=ES.tx_ac
            where tx_ac=? and alt_ac=? and alt_aln_method=?
            """,

        "tx_mapping_options": """
            select distinct tx_ac,alt_ac,alt_aln_method
            from tx_exon_aln_v where tx_ac=? and exon_aln_id is not NULL
            """,

        "tx_seq": """
            select seq
            from seq S
            join seq_anno SA on S.seq_id=SA.seq_id
            where ac=?
            """,
    }

    def __init__(self):
        self._check_schema_version('1')


    def data_version(self):
        cur = self._get_cursor()
        cur.execute("select * from meta where key = 'schema_version'")
        return cur.fetchone()['value']


    def schema_version(self):
        return self.data_version()


    ############################################################################
    ## Queries
    def get_acs_for_protein_seq(self,seq):
        """
        returns a list of protein accessions for a given sequence.  The
        list is guaranteed to contain at least one element with the
        MD5-based accession (MD5_01234abc...def56789) at the end of the
        list.
        """
        md5 = seq_md5(seq)
        cur = self._get_cursor()
        cur.execute(self.sql['acs_for_protein_md5'],[md5])
        return [ r['ac'] for r in cur.fetchall() ] + [ 'MD5_'+md5 ]


    def get_gene_info(self, gene):
        """
        returns basic information about the gene.

        :param gene: HGNC gene name
        :type gene: str

        # database results
        hgnc    | ATM
        maploc  | 11q22-q23
        descr   | ataxia telangiectasia mutated
        summary | The protein encoded by this gene belongs to the PI3/PI4-kinase family. This...
        aliases | AT1,ATA,ATC,ATD,ATE,ATDC,TEL1,TELO1
        added   | 2014-02-04 21:39:32.57125

        """
        cur = self._get_cursor()
        cur.execute(self.sql['gene_info'],[gene])
        return cur.fetchone()


    def get_tx_exons(self, tx_ac, alt_ac, alt_aln_method):
        """
        return transcript exon info for supplied accession (tx_ac, alt_ac, alt_aln_method), or None if not found
        
        :param tx_ac: transcript accession with version (e.g., 'NM_000051.3')
        :type tx_ac: str

        :param alt_ac: specific genomic sequence (e.g., NC_000011.4)
        :type alt_ac: str

        :param alt_aln_method: sequence alignment method (e.g., splign, blat)
        :type alt_aln_method: str
        
        # tx_exons = db.get_tx_exons('NM_199425.2', 'NC_000020.10', 'splign')
        # len(tx_exons)
        3
        
        tx_exons have the following attributes::
        
            {
                'tes_exon_set_id' : 98390
                'aes_exon_set_id' : 298679
                'tx_ac'           : 'NM_199425.2'
                'alt_ac'          : 'NC_000020.10'
                'alt_strand'      : -1
                'alt_aln_method'  : 'splign'
                'ord'             : 2
                'tx_exon_id'      : 936834
                'alt_exon_id'     : 2999028
                'tx_start_i'      : 786
                'tx_end_i'        : 1196
                'alt_start_i'     : 25059178
                'alt_end_i'       : 25059588
                'cigar'           : '410='
            }

        For example:
        
        # tx_exons[0]['tx_ac']
        'NM_199425.2'
        
        """
        cur = self._get_cursor()
        cur.execute(self.sql['tx_exons'],[tx_ac, alt_ac, alt_aln_method])
        r = cur.fetchall()
        if len(r) == 0:
            return None
        else:
            return r


    def get_tx_for_gene(self,gene):
        """
        return transcript info records for supplied gene, in order of decreasing length

        :param gene: HGNC gene name
        :type gene: str
        """
        cur = self._get_cursor()
        cur.execute(self.sql['tx_for_gene'],[gene])
        return cur.fetchall()

    def get_tx_for_region(self,alt_ac,alt_aln_method,start_i,end_i):
        """
        return transcripts that overlap given region

        :param str alt_ac: reference sequence (e.g., NC_000007.13)
        :param str alt_aln_method: alignment method (e.g., splign)
        :param int start_i: 5' bound of region
        :param int end_i: 3' bound of region
        """
        cur = self._get_cursor()
        cur.execute(self.sql['tx_for_region'],[alt_ac,alt_aln_method,start_i,end_i])
        return cur.fetchall()

    def get_tx_identity_info(self, tx_ac):
        """returns features associated with a single transcript.

        :param tx_ac: transcript accession with version (e.g., 'NM_199425.2')
        :type tx_ac: str

        # database output
        -[ RECORD 1 ]--+-------------
        tx_ac          | NM_199425.2
        alt_ac         | NM_199425.2
        alt_aln_method | transcript
        cds_start_i    | 283
        cds_end_i      | 1003
        lengths        | {707,79,410}
        hgnc           | VSX1

        """
        cur = self._get_cursor()
        cur.execute(self.sql['tx_identity_info'],[tx_ac])
        return cur.fetchone()


    def get_tx_info(self, tx_ac, alt_ac, alt_aln_method):
        """return a single transcript info for supplied accession (tx_ac, alt_ac, alt_aln_method), or None if not found

        :param tx_ac: transcript accession with version (e.g., 'NM_000051.3')
        :type tx_ac: str

        :param alt_ac: specific genomic sequence (e.g., NC_000011.4)
        :type alt_ac: str

        :param alt_aln_method: sequence alignment method (e.g., splign, blat)
        :type alt_aln_method: str

        # database output
        -[ RECORD 1 ]--+------------
        hgnc           | ATM
        cds_start_i    | 385
        cds_end_i      | 9556
        tx_ac          | NM_000051.3
        alt_ac         | AC_000143.1
        alt_aln_method | splign

        """
        cur = self._get_cursor()
        cur.execute(self.sql['tx_info'],[tx_ac, alt_ac, alt_aln_method])
        return cur.fetchone()


    def get_tx_seq(self,ac):
        """return transcript sequence for supplied accession (ac), or None if not found

        :param ac: transcript accession with version (e.g., 'NM_000051.3')
        :type ac: str
        """
        cur = self._get_cursor()
        cur.execute(self.sql['tx_seq'],[ac])
        try:
            return cur.fetchone()['seq']
        except TypeError:
            return None


    def get_tx_mapping_options(self, tx_ac):
        """Return all transcript alignment sets for a given transcript
        accession (tx_ac); returns empty list if transcript does not
        exist.  Use this method to discovery possible mapping options
        supported in the database

        :param tx_ac: transcript accession with version (e.g., 'NM_000051.3')
        :type tx_ac: str

        # database output
        -[ RECORD 1 ]--+------------
        hgnc           | ATM
        cds_start_i    | 385
        cds_end_i      | 9556
        tx_ac          | NM_000051.3
        alt_ac         | AC_000143.1
        alt_aln_method | splign
        -[ RECORD 2 ]--+------------
        hgnc           | ATM
        cds_start_i    | 385
        cds_end_i      | 9556
        tx_ac          | NM_000051.3
        alt_ac         | NC_000011.9
        alt_aln_method | blat

        """
        cur = self._get_cursor()
        cur.execute(self.sql['tx_mapping_options'],[tx_ac])
        r = cur.fetchall()
        return r




    ############################################################################
    ## INTERNAL FUNCTIONS
    def _check_schema_version(self,required_version):
        obs_Mm = self.schema_version().split('.')[:2]
        req_Mm = required_version.split('.')[:2]
        if ( obs_Mm != req_Mm ):
            raise RuntimeError("Version mismatch: Version {req_Mm} required, but {self.db_url} is version {obs_Mm}".format(
                req_Mm = '.'.join(req_Mm), self=self, obs_Mm = '.'.join(obs_Mm)))
        # else no error



class UTA_sqlite(UTABase):
    def __init__(self,url):
        def _sqlite3_row_dict_factory(cur, row):
            "convert sqlite row to dict"
            return dict( (d[0],row[i]) for i,d in enumerate(cur.description) )

        self.db_url = url.geturl()

        db_path = url.path
        if not os.path.exists(db_path):
            raise IOError(db_path + ': Non-existent database file')
        self._con = sqlite3.connect(db_path)
        self._con.row_factory = _sqlite3_row_dict_factory
        logger.info("connected to "+db_path)
        super(UTA_sqlite,self).__init__()
        
    def _get_cursor(self):
        return self._con.cursor()


class UTA_postgresql(UTABase):
    def __init__(self,url):
        self.db_url = url.geturl()

        host = url.hostname
        port = url.port or 5432
        database = url.path[1:]
        user = url.username
        password= url.password

        logger.info('connecting to '+self.db_url)
        logger.debug('connection params: host={host}, port={port}, database={database}, user={user}, password={pw}'.format(
            host=host,port=port,database=database,user=user,pw='' if password is None else '***'))
        self._con = psycopg2.connect(host=host, port=port, database=database, user=user, password=password)
        super(UTA_postgresql,self).__init__()

        # remap sqlite's ? placeholders to psycopg2's %s
        self.sql = { k: v.replace('?','%s') for k,v in self.sql.iteritems() }
        
    def _get_cursor(self):
        cur = self._con.cursor(cursor_factory=psycopg2.extras.DictCursor)
        cur.execute('set search_path = '+pg_schema)
        return cur




## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



import unittest

from nose.plugins.attrib import attr

import hgvs.utils

@attr(tags=["quick"])
class Test_Utils(unittest.TestCase):
    # set of aa1 and aa3 -- must be in alpha order
    all_aa1 = '*ACDEFGHIKLMNPQRSTUVWXY'
    all_aa3 = 'AlaArgAsnAspCysGlnGluGlyHisIleLeuLysMetPheProSecSerTerThrTrpTyrValXaa'

    # sequence of aa1 and aa3 in corresponding order
    aa1_seq = 'YWVTSRQPNMLKIHGFEDCAXU*'
    aa3_seq = 'TyrTrpValThrSerArgGlnProAsnMetLeuLysIleHisGlyPheGluAspCysAlaXaaSecTer'


    def test_aa1_lut(self):
        self.assertEqual( len(hgvs.utils.aa1_to_aa3_lut), 23 )
        self.assertEqual( ''.join(sorted(hgvs.utils.aa1_to_aa3_lut.keys())),   self.all_aa1 )
        self.assertEqual( ''.join(sorted(hgvs.utils.aa1_to_aa3_lut.values())), self.all_aa3 )

    def test_aa3_lut(self):
        self.assertEqual( len(hgvs.utils.aa3_to_aa1_lut), 23 )
        self.assertEqual( ''.join(sorted(hgvs.utils.aa3_to_aa1_lut.values())), self.all_aa1 )
        self.assertEqual( ''.join(sorted(hgvs.utils.aa3_to_aa1_lut.keys())),   self.all_aa3 )


    def test_aa1_to_aa3(self):
        self.assertEqual( hgvs.utils.aa3_to_aa1(self.aa3_seq), self.aa1_seq )

    def test_aa3_to_aa1(self):
        self.assertEqual( hgvs.utils.aa1_to_aa3(self.aa1_seq), self.aa3_seq )


    def test_aa_to_aa1(self):
        self.assertEqual( hgvs.utils.aa_to_aa1('')		, ''   )
        self.assertEqual( hgvs.utils.aa_to_aa1('A')     , 'A'  )
        self.assertEqual( hgvs.utils.aa_to_aa1('AC')    , 'AC' )
        self.assertEqual( hgvs.utils.aa_to_aa1('Ala')   , 'A'  )
        self.assertEqual( hgvs.utils.aa_to_aa1('AlaCys'), 'AC' )
        self.assertEqual( hgvs.utils.aa_to_aa1(self.aa1_seq), self.aa1_seq )
        self.assertEqual( hgvs.utils.aa_to_aa1(self.aa3_seq), self.aa1_seq )

    def test_aa_to_aa3(self):
        self.assertEqual( hgvs.utils.aa_to_aa3('')      , ''       )
        self.assertEqual( hgvs.utils.aa_to_aa3('A')     , 'Ala'    )
        self.assertEqual( hgvs.utils.aa_to_aa3('AC')    , 'AlaCys' )
        self.assertEqual( hgvs.utils.aa_to_aa3('Ala')   , 'Ala'    )
        self.assertEqual( hgvs.utils.aa_to_aa3('AlaCys'), 'AlaCys' )
        self.assertEqual( hgvs.utils.aa_to_aa3(self.aa1_seq), self.aa3_seq )
        self.assertEqual( hgvs.utils.aa_to_aa3(self.aa3_seq), self.aa3_seq )


if __name__ == '__main__':
    unittest.main()


## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



__doc__ = """
hgvs.edit -- representation of edit operations in HGVS variants

NARefAlt and AARefAlt are abstractions of several major variant
types.  They are distinguished by whether the ref and alt elements
of the structure.  The HGVS grammar for NA and AA are subtly
different (e.g., the ref AA in a protein substitution is part of the
location).

"""


import recordtype

from hgvs.exceptions import HGVSError
from hgvs.utils import aa_to_aa1,aa1_to_aa3


class Edit(object):
    pass


class NARefAlt( Edit, recordtype.recordtype('NARefAlt', [('ref',None),('alt',None),('uncertain',False)]) ):
    """
    represents substitutions, deletions, insertions, and indels.

    :ivar ref: reference sequence or length
    :ivar alt: alternate sequence
    :ivar uncertain: boolean indicating whether the variant is uncertain/undetermined
    """

    @property
    def ref_s(self):
        """
        returns a string representing the ref sequence, if it is not None and smells like a sequence

        >>> NARefAlt('ACGT').ref_s
        'ACGT'
        >>> NARefAlt('7').ref_s
        >>> NARefAlt(7).ref_s

        """
        return self.ref if (isinstance(self.ref,basestring) and self.ref and self.ref[0] in 'ACGTUN') else None

    @property
    def ref_n(self):
        """
        returns an integer, either from the `ref` instance variable if it's a number, or the length of
        ref if it's a string, or None otherwise

        >>> NARefAlt('ACGT').ref_n
        4
        >>> NARefAlt('7').ref_n
        7
        >>> NARefAlt(7).ref_n
        7

        """
        try:
            return int(self.ref)
        except ValueError:
            return len(self.ref) if self.ref else None


    def __str__(self):
        if self.ref is None and self.alt is None:
            raise HGVSError('RefAlt: ref and alt sequences are both undefined')

        # subst and delins
        if self.ref is not None and self.alt is not None:
            if self.ref == self.alt:
                s = '='
            elif len(self.alt) == 1 and len(self.ref) == 1 and not self.ref.isdigit(): # don't turn del5insT into 5>T
                s = '{self.ref}>{self.alt}'.format(self=self)
            else:
                s = 'del{self.ref}ins{self.alt}'.format(self=self)
        # del case
        elif self.ref is not None:
            s = 'del{self.ref}'.format(self=self)

        # ins case
        else:   # self.alt is not None
            s = 'ins{self.alt}'.format(self=self)

        return '('+s+')' if self.uncertain else s

    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        if self.ref is not None and self.alt is not None:
            if self.ref == self.alt:
                edit_type = 'identity'
            elif len(self.alt) == 1 and len(self.ref) == 1 and not self.ref.isdigit():
                edit_type = 'sub'
            else:
                edit_type = 'delins'
        elif self.ref is not None:
            edit_type = 'del'
        else:
            edit_type = 'ins'
        return edit_type


class AARefAlt( Edit, recordtype.recordtype('AARefAlt', [('ref',None),('alt',None), ('uncertain',False)]) ):
    def __init__(self,ref, alt, uncertain=False):
        super(AARefAlt, self).__init__(ref=aa_to_aa1(ref), alt=aa_to_aa1(alt), uncertain=uncertain)

    def __str__(self):
        if self.ref is None and self.alt is None:
            #raise HGVSError('RefAlt: ref and alt sequences are both undefined')
            return '='

        # subst and delins
        if self.ref is not None and self.alt is not None:
            if self.ref == self.alt:
                s = '='
            elif len(self.ref) == 1 and len(self.alt) == 1:
                s = aa1_to_aa3(self.alt)
            else:
                s = 'delins{alt}'.format(alt = aa1_to_aa3(self.alt))

        # del case
        elif self.ref is not None and self.alt is None:
            s = 'del'

        # ins case
        elif self.ref is None and self.alt is not None:
            s = 'ins{alt}'.format(alt=aa1_to_aa3(self.alt))

        else:
            raise RuntimeError("Should not be here")
            
        return '('+s+')' if self.uncertain else s
    
    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        if self.ref is not None and self.alt is not None:
            if self.ref == self.alt:
                edit_type = 'identity'
            elif len(self.ref) == 1 and len(self.alt) == 1:
                edit_type = 'sub'
            else:
                edit_type = 'delins'
        elif self.ref is not None and self.alt is None:
            edit_type = 'del'
        elif self.ref is None and self.alt is not None:
            edit_type = 'ins'
        return edit_type


class AASub( AARefAlt ):
    def __str__(self):
        s = aa1_to_aa3(self.alt) if self.alt != '?' else self.alt
        return '('+s+')' if self.uncertain else s

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        return 'sub'


class AAFs(Edit, recordtype.recordtype('AAFs', [('ref',None),('alt',None),('length',None),('uncertain',False)])):
    def __init__(self,ref,alt,length=None,uncertain=False):
        super(AAFs, self).__init__(ref=aa_to_aa1(ref), alt=aa_to_aa1(alt), length=length, uncertain=uncertain)

    def __str__(self):
        st_length = self.length or ''
        s = "{alt}fsTer{length}".format(alt=aa1_to_aa3(self.alt), length=st_length)
        return '('+s+')' if self.uncertain else s

    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        return 'fs'


class AAExt(Edit, recordtype.recordtype('AAExt', [('ref',None),('alt',None), ('aaterm', None), ('length',None),
                                                  ('uncertain',False)])):
    def __init__(self,ref,alt,aaterm=None, length=None,uncertain=False):
        super(AAExt, self).__init__(ref=aa_to_aa1(ref), alt=aa_to_aa1(alt), aaterm=aa_to_aa1(aaterm), length=length,
                                    uncertain=uncertain)

    def __str__(self):
        st_alt = self.alt or ''
        st_aaterm = self.aaterm or ''
        st_length = self.length or ''
        s = "{alt}ext{term}{length}".format(alt=aa1_to_aa3(st_alt), term=aa1_to_aa3(st_aaterm), length=st_length)
        return '('+s+')' if self.uncertain else s

    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        return 'ext'


class Dup( Edit, recordtype.recordtype('Dup', [('seq',None),('uncertain',False)]) ):

    def __str__(self):
        return 'dup' + (self.seq or '')

    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        return 'dup'


class Repeat( Edit, recordtype.recordtype('Repeat', [('seq',None),('min',None),('max',None),('uncertain',False)]) ):

    def __str__(self):
        if self.min > self.max:
            raise HGVSError('Repeat min count must be less than or equal to max count')
        if self.min == self.max:
            return '{self.seq}[{self.min}]'.format(self=self)
        return '{self.seq}({self.min}_{self.max})'.format(self=self)

    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        return 'repeat'


class NACopy(Edit, recordtype.recordtype('NACopy', ['copy', ('uncertain', False)])):

    def __str__(self):
        s = 'copy{}'.format(self.copy)
        return '('+s+')' if self.uncertain else s

    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        return 'copy'


class NADupN(Edit, recordtype.recordtype('NADupN', ['n', ('uncertain', False)])):

    def __str__(self):
        s = 'dup{}'.format(self.n)
        return '('+s+')' if self.uncertain else s

    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        return 'dup'


if __name__ == "__main__":
    import doctest
    doctest.testmod()


# class Inv( Edit, recordtype.recordtype('Inv', [], default=None) ):
#     def __str__(self):
#         return ''
# 
# class Con( Edit, recordtype.recordtype('Con', ['con'], default=None) ):
#     def __str__(self):
#         return self.con
# 
# class ComplexVariant( Edit, recordtype.recordtype('ComplexVariant', ['edits','rel'], default=None) ):
#     def __str__(self):
#         return '[' + self.rel.join( self.edits ) + ']'
# 
# class CompoundVariant( Edit, recordtype.recordtype('CompoundVariant', ['edits'], default=None) ):
#     def __str__(self):
#         return ';'.join( [ '['+e+']' for e in self.edits ] )
# 
# class MosaicVariant( Edit, recordtype.recordtype('Edit', ['edit'], default=None) ):
#     def __str__(self):
#         return '[=/{self.edit}]'.format(self=self)
# 
# class ChimericVariant( Edit, recordtype.recordtype('Edit', ['edit'], default=None) ):
#     def __str__(self):
#         return '[=//{self.edit}]'.format(self=self)


## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- encoding: utf-8 -*-

import recordtype

class HGVSPosition( recordtype.recordtype('HGVSPosition', ['ac','type','pos']) ):
    """
    HGVSPosition -- Represent partial HGVS tags that refer to a position without alleles
    
    :param str ac: sequence accession
    :param str type: type of sequence and coordinate
    :param str pos: sequence position
    
    """
    def __str__(self):
        return '{self.ac}:{self.type}.{self.pos}'.format(self=self)


## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# we're not testing hgvs.parser, but rather merely using it to load the
# grammar.  See test_hgvs_parser.py for the parser tests

import unittest

from nose.plugins.attrib import attr

import hgvs.parser

@attr(tags=["quick"])
class Test_Parser(unittest.TestCase):
    longMessage = True

    def setUp(self):
        self.p = hgvs.parser.Parser()
        self.grammar = self.p._grammar

    def test_parser_basic(self):
        # num & snum
        self.assertEqual( self.grammar('55').num(), 55 )
        self.assertEqual( self.grammar('55').snum(), 55 )
        self.assertEqual( self.grammar('-55').snum(), -55 )

        # accn
        self.assertEqual( self.grammar('NM_01234.5').accn(), 'NM_01234.5' )
        self.assertEqual( self.grammar('NM_01234').accn(), 'NM_01234' )
        self.assertEqual( self.grammar('ENST01234').accn(), 'ENST01234' )
        self.assertEqual( self.grammar('LRG_1234').accn(), 'LRG_1234' )
        self.assertEqual( self.grammar('LRG_1234t1').accn(), 'LRG_1234t1' )

    def test_parser_aa(self):
        # terms
        self.assertEqual( self.p.parse_term1('*'), '*' )
        self.assertEqual( self.p.parse_term3('Ter'), 'Ter' )
        self.assertEqual( self.p.parse_term13('*'), '*' )
        self.assertEqual( self.p.parse_term13('Ter'), 'Ter' )

        # AAs
        self.assertEqual( self.p.parse_aa1('A'), 'A' )
        self.assertEqual( self.p.parse_aa3('Ala'), 'Ala' )
        self.assertEqual( self.p.parse_aa13('A'), 'A' )
        self.assertEqual( self.p.parse_aa13('Ala'), 'Ala' )

        # mixto
        self.assertEqual( self.p.parse_aat1('*'), '*' )
        self.assertEqual( self.p.parse_aat1('A'), 'A' )
        self.assertEqual( self.p.parse_aat3('Ter'), 'Ter' )
        self.assertEqual( self.p.parse_aat3('Ala'), 'Ala' )
        self.assertEqual( self.p.parse_aat13('*'), '*' )
        self.assertEqual( self.p.parse_aat13('A'), 'A' )
        self.assertEqual( self.p.parse_aat13('Ter'), 'Ter' )
        self.assertEqual( self.p.parse_aat13('Ala'), 'Ala' )

    def test_parser_aa_seq(self):
        # mixto
        self.assertEqual( self.p.parse_aat1_seq('*'), '*' )
        self.assertEqual( self.p.parse_aat1_seq('A'), 'A' )
        self.assertEqual( self.p.parse_aat3_seq('Ter'), 'Ter' )
        self.assertEqual( self.p.parse_aat3_seq('Ala'), 'Ala' )
        self.assertEqual( self.p.parse_aat13_seq('*'), '*' )
        self.assertEqual( self.p.parse_aat13_seq('A'), 'A' )
        self.assertEqual( self.p.parse_aat13_seq('Ter'), 'Ter' )
        self.assertEqual( self.p.parse_aat13_seq('Ala'), 'Ala' )



if __name__ == '__main__':
    unittest.main()

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



#
# Tests of the grammar
#
# Code takes a tab-delimited text file of the form:
#
#   Func    Test    Valid   InType  Expected
#   pm      -+      True    string
#   pm      *       False   one
#   num     1|+1    True    list    1|1
#
# Headers are defined as follows:
# Func: function name to call in the grammar
# Test: item(s) to test
# Valid: if the input is expected to be valid (True or False)
# InType: 3 type:
# - one: input is a single value
# - string: input is a string; test each character in the string separately
# - list: input is a list delimited by a pipe character ('|')
# Expected: expected result (if stringifying input does not return the same answer, e,g. '+1' -> '1')
# - if expected is left blank, then it is assumed that stringifying the parsed input returns the same answer.
#
from __future__ import with_statement

import os
import csv
import pprint
import unittest

import pkg_resources

import hgvs.parser

class TestGrammarFull(unittest.TestCase):

    def setUp(self):
        self.p = hgvs.parser.Parser()
        self.grammar = self.p._grammar
        self._test_fn = os.path.join(os.path.dirname(__file__), 'data', 'grammar_test.tsv')

    def test_parser_test_completeness(self):

        rules_tested = set()

        with open(self._test_fn, 'r') as f:
            reader = csv.DictReader(f, delimiter='\t')
            for row in reader:
                rules_tested.add(row['Func'])

        rules_all= set()
        grammar_fn = pkg_resources.resource_filename(__name__, '../hgvs/_data/hgvs.pymeta')
        for line in open(grammar_fn, 'r'):
            if len(line) > 0 and line[0] != '#' and line[0].isalpha():
                line = line.strip()
                rules_all.add(line.split()[0])

        rules_untested = rules_all - rules_tested

        msg = "untested rules: {}".format(rules_untested)
        self.assertTrue(len(rules_untested) == 0, msg)

    def test_parser_grammar(self):
        with open(self._test_fn, 'r') as f:
            reader = csv.DictReader(f, delimiter='\t')

            fail_cases = []

            for row in reader:
                if row['Func'].startswith('#'):
                    continue

                # setup input
                inputs = self._split_inputs(row['Test'], row['InType'])
                expected_results = self._split_inputs(row['Expected'], row['InType']) \
                    if row['Expected'] else inputs
                expected_map = dict(zip(inputs, expected_results))

                # step through each item and check
                is_valid = True if row['Valid'].lower() == 'true' else False

                for key in expected_map:
                    expected_result = str(expected_map[key])
                    function_to_test = getattr(self.p._grammar(key), row['Func'])
                    row_str = "{}\t{}\t{}\t{}\t{}".format(row['Func'], key, row['Valid'], 'one', expected_result)
                    try:
                        actual_result = str(function_to_test())
                        if not is_valid or (expected_result != actual_result):
                            print "expected: {} actual:{}".format(expected_result, actual_result)
                            fail_cases.append(row_str)
                    except Exception as e:
                        if is_valid:
                            print "expected: {} Exception: {}".format(expected_result, e)
                            fail_cases.append(row_str)

        # everything should have passed - report whatever failed
        self.assertTrue(len(fail_cases) == 0, pprint.pprint(fail_cases))


    def _split_inputs(self, in_string, intype):
        DELIM = '|'
        if intype == 'list':
            inputs = in_string.split(DELIM)
        elif intype == 'string':
            inputs = list(in_string)
        else:   # intype == 'one'
            inputs = [in_string]
        inputs = [x if x != 'None' else None for x in inputs]
        return inputs

if __name__ == '__main__':
    unittest.main()

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



import os
import unittest

import hgvs.dataproviders.multifastadb

# ==> d1/f1.fasta <==
# >s1
# d1f1s1
# >s2
# d1f1s2
# 
# ==> d1/f2.fasta <==
# >s3
# d1f2s3
# >s4
# d1f2s4
# 
# ==> d2/f3.fa <==
# >s1
# d2f3s1
# >s5
# d2f3s5
# 
# ==> d2/f4.fa <==
# >s3
# d2f4s3
# >s6
# d2f4s6


class MFDBTBase(unittest.TestCase):
    test_dir = os.path.join(os.path.dirname(__file__),'data','multifastadb')
    print test_dir
    def setUp(self):
        self.mfdb = hgvs.dataproviders.multifastadb.MultiFastaDB([ os.path.join(self.test_dir,s)
                                                     for s in self.sources ], use_meta_index=self.use_meta_index)


class MFDB_Test_d1f1(MFDBTBase):
    sources = ['d1/f1.fasta']
    use_meta_index = False

    def test_basic_file(self):
        self.assertEqual( self.mfdb.references, ['s1','s2'] )
        self.assertEqual( self.mfdb.lengths, [6,6] )

    def test_basic_sequence(self):
        self.assertEqual( self.mfdb.fetch('s1'), 'd1f1s1' )
        self.assertEqual( str(self.mfdb['s1']), 'd1f1s1' )
        self.assertEqual( self.mfdb['s1'][:2], 'd1' )


class MFDB_Test_d1(MFDBTBase):
    sources = ['d1']
    use_meta_index = False

    def test_basic_file(self):
        self.assertEqual( self.mfdb.references, ['s1','s2','s3','s4'] )
        self.assertEqual( self.mfdb.lengths, [6,6,6,6] )

    def test_basic_sequence(self):
        self.assertEqual( self.mfdb.fetch('s1'), 'd1f1s1' )
        self.assertEqual( str(self.mfdb['s1']), 'd1f1s1' )
        self.assertEqual( self.mfdb['s1'][:2], 'd1' )
        self.assertEqual( self.mfdb.fetch('s3'), 'd1f2s3' )
        self.assertEqual( str(self.mfdb['s3']), 'd1f2s3' )
        self.assertEqual( self.mfdb['s3'][:2], 'd1' )

class MFDB_Test_d2d1(MFDBTBase):
    sources = ['d2','d1']
    use_meta_index = False

    def test_basic_file(self):
        self.assertEqual( self.mfdb.references, ['s1','s5','s3','s6','s1','s2','s3','s4'] )
        self.assertEqual( self.mfdb.lengths, [6,6,6,6,6,6,6,6] )

    def test_basic_sequence(self):
        # s1 appears twice; should get d2f3 version
        self.assertEqual( self.mfdb.fetch('s1'), 'd2f3s1' )
        self.assertEqual( str(self.mfdb['s1']), 'd2f3s1' )
        self.assertEqual( self.mfdb['s1'][:2], 'd2' )

        self.assertEqual( self.mfdb.fetch('s5'), 'd2f3s5' )
        self.assertEqual( str(self.mfdb['s5']), 'd2f3s5' )
        self.assertEqual( self.mfdb['s5'][:2], 'd2' )

class MFDB_Test_Actual(MFDBTBase):
    sources = ['real_examples']
    use_meta_index = True

    def test_basic_file(self):
        self.assertEquals( self.mfdb.references, ['gi|53292629|ref|NP_001005405.1|',
                                                  'gi|52317162|ref|NP_001004713.1|',
                                                  'gi|123173798|ref|NM_001005405.2|',
                                                  'gi|52317161|ref|NM_001004713.1|',
                                                  'gi|296923737|ref|NG_021245.2|',
                                                  'gi|226510210|ref|NG_011805.1|',
                                                  ])
        self.assertEquals( self.mfdb.lengths, [156,355,1029,1068,96420,1169911] )

    def test_basic_sequence(self):
        self.assertEquals( self.mfdb.fetch('gi|53292629|ref|NP_001005405.1|'), 'MGCCGCSGGCGSGCGGCGSGSGGCGSGCGGCGSSCCVPICCCKPVCCCVPACSCSSCGSCGGSKGGCGSCGSSKGGCGSCGCSQSNCCKPCCSSSGCGSFCCQSSCSKPCCCQSSCCQSSCCKPCCCQSSCCQSSCFKPCCCQSSCCVPVCCQCKI')
        self.assertEquals( self.mfdb.fetch('NP_001005405.1'), 'MGCCGCSGGCGSGCGGCGSGSGGCGSGCGGCGSSCCVPICCCKPVCCCVPACSCSSCGSCGGSKGGCGSCGSSKGGCGSCGCSQSNCCKPCCSSSGCGSFCCQSSCSKPCCCQSSCCQSSCCKPCCCQSSCCQSSCFKPCCCQSSCCVPVCCQCKI')
        self.assertEquals( self.mfdb.fetch('NP_001005405.1',0,3), 'MGC')

if __name__ == '__main__':
    unittest.main()

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



import math
import re

import hgvs.location
import hgvs.variant
import hgvs.posedit
from hgvs.intervalmapper import IntervalMapper
from hgvs.exceptions import HGVSError
from hgvs.utils.deprecated import deprecated

class TranscriptMapper(object):
    """Provides coordinate (not variant) mapping operations between
    genomic (g), rna (r), cds (c), and protein (p) coordinates.  All
    coordinates are 1-based inclusive, per the HGVS recommendations.
    All methods take :class:`hgvs.location.Interval` objects.

    :param hdp: HGVS Data Provider Interface-compliant instance (see :class:`hgvs.dataproviders.interface.Interface`)
    :param str tx_ac: string representing transcript accession (e.g., NM_000551.2)
    :param str alt_ac: string representing the reference sequence accession (e.g., NM_000551.3)
    :param str alt_aln_method: string representing the alignment method; valid values depend on data source

    """

    def __init__(self, hdp, tx_ac, alt_ac, alt_aln_method):
        self.hdp = hdp
        self.tx_ac = tx_ac
        self.alt_ac = alt_ac
        self.alt_aln_method = alt_aln_method
        if self.alt_aln_method != 'transcript':
            self.tx_info = hdp.get_tx_info(self.tx_ac, self.alt_ac, self.alt_aln_method)
            self.tx_exons = hdp.get_tx_exons(self.tx_ac, self.alt_ac, self.alt_aln_method)
            if self.tx_info is None or self.tx_exons is None:
                raise HGVSError("Couldn't build TranscriptMapper(tx_ac={self.tx_ac}, "
                                "alt_ac={self.alt_ac}, alt_aln_method={self.alt_aln_method})".format(self=self))
            self.strand = self.tx_exons[0]['alt_strand']
            self.cds_start_i = self.tx_info['cds_start_i']
            self.cds_end_i = self.tx_info['cds_end_i']
            self.gc_offset = self.tx_exons[0]['alt_start_i']
            self.cigar = _build_tx_cigar(self.tx_exons, self.strand)
            self.im = IntervalMapper.from_cigar(self.cigar)
            self.tgt_len = self.im.tgt_len
        else:
            # this covers the identity cases r <-> c
            self.tx_identity_info = hdp.get_tx_identity_info(self.tx_ac)
            self.cds_start_i = self.tx_identity_info['cds_start_i']
            self.cds_end_i = self.tx_identity_info['cds_end_i']
            self.tgt_len = sum(self.tx_identity_info['lengths'])


    def __str__(self):
        return '{self.__class__.__name__}: {self.tx_ac} ~ {self.alt_ac} ~ {self.alt_aln_method); ' \
               '{strand_pm} strand; {n_exons} exons; offset={self.gc_offset}'.format(
                   self=self, n_exons=len(self.tx_exons), strand_pm=_strand_pm(self.strand))


    def g_to_r(self, g_interval):
        """convert a genomic (g.) interval to a transcript cDNA (r.) interval"""

        assert self.strand in [1,-1], 'strand = '+str(self.strand)+'; must be 1 or -1'

        g_ci = _hgvs_coord_to_ci(g_interval.start.base, g_interval.end.base)
        # frs, fre = (f)orward (r)na (s)tart & (e)nd; forward w.r.t. genome
        frs, fre = self.im.map_ref_to_tgt(g_ci[0] - self.gc_offset, g_ci[1] - self.gc_offset, max_extent=False)
        if self.strand == -1:
            frs, fre = self.tgt_len - fre, self.tgt_len - frs
        # get BaseOffsetPosition information for r.
        if self.strand == 1:
            grs, gre = self.im.map_tgt_to_ref(frs, fre, max_extent=False)
        elif self.strand == -1:
            grs, gre = self.im.map_tgt_to_ref((self.tgt_len - fre), (self.tgt_len - frs), max_extent=False)
        grs, gre = grs + self.gc_offset, gre + self.gc_offset
        # 0-width interval indicates an intron.  Need to calculate offsets but we're are in ci coordinates
        # requires adding 1 strategically to get the HGVS position (shift coordinates to 3' end of the ref nucleotide)
        if frs == fre:
            start_offset = _hgvs_offset(g_ci[0] + 1, grs, gre + 1)
            end_offset = _hgvs_offset(g_ci[1], grs, gre + 1)
        else:
            start_offset = _hgvs_offset(g_ci[0], grs, gre)
            end_offset = _hgvs_offset(g_ci[1], grs, gre)
        if self.strand == -1:
            start_offset, end_offset = self.strand * end_offset, self.strand * start_offset
        if start_offset > 0:
            frs -= 1
        if end_offset < 0:
            fre += 1
        return hgvs.location.Interval(
                    start=hgvs.location.BaseOffsetPosition(base=_ci_to_hgvs_coord(frs, fre)[0], offset=start_offset),
                    end  =hgvs.location.BaseOffsetPosition(base=_ci_to_hgvs_coord(frs, fre)[1], offset=end_offset),
                    uncertain=g_interval.uncertain)


    def r_to_g(self, r_interval):
        """convert a transcript cDNA (r.) interval to a genomic (g.) interval"""

        assert self.strand in [1,-1], 'strand = '+str(self.strand)+'; must be 1 or -1'

        if self.strand == 1:
            frs, fre = _hgvs_coord_to_ci(r_interval.start.base, r_interval.end.base)
            start_offset, end_offset = r_interval.start.offset, r_interval.end.offset
        elif self.strand == -1:
            frs, fre = _hgvs_coord_to_ci(r_interval.start.base, r_interval.end.base)
            fre, frs = self.tgt_len - frs, self.tgt_len - fre
            start_offset, end_offset = self.strand * r_interval.end.offset, self.strand * r_interval.start.offset

        # returns the genomic range start (grs) and end (gre)
        grs, gre = self.im.map_tgt_to_ref(frs, fre, max_extent=False)
        grs, gre = grs + self.gc_offset, gre + self.gc_offset
        gs, ge = grs + start_offset, gre + end_offset
        return hgvs.location.Interval(
            start=hgvs.location.SimplePosition(_ci_to_hgvs_coord(gs, ge)[0], uncertain=r_interval.start.uncertain),
            end  =hgvs.location.SimplePosition(_ci_to_hgvs_coord(gs, ge)[1], uncertain=r_interval.end.uncertain),
            uncertain=r_interval.uncertain)


    def r_to_c(self, r_interval):
        """convert a transcript cDNA (r.) interval to a transcript CDS (c.) interval"""

        if r_interval.start.base <= 0:
            raise HGVSError("Transcript out of bounds.  Start position ({rs}) is <= 0.".
                            format(rs=r_interval.start.base))
        if r_interval.end.base > self.tgt_len:
            raise HGVSError("Transcript out of bounds. End position ({re}) is > than transcript length ({len}).".
                            format(re=r_interval.end.base, len=self.tgt_len))
        # start
        if r_interval.start.base <= self.cds_start_i:
            cs = r_interval.start.base - (self.cds_start_i + 1)
            cs_datum = hgvs.location.CDS_START
        elif r_interval.start.base > self.cds_start_i and r_interval.start.base <= self.cds_end_i:
            cs = r_interval.start.base - self.cds_start_i
            cs_datum = hgvs.location.CDS_START
        else:
            cs = r_interval.start.base - self.cds_end_i
            cs_datum = hgvs.location.CDS_END
        # end
        if r_interval.end.base <= self.cds_start_i:
            ce = r_interval.end.base - (self.cds_start_i + 1)
            ce_datum = hgvs.location.CDS_START
        elif r_interval.end.base > self.cds_start_i and r_interval.end.base <= self.cds_end_i:
            ce = r_interval.end.base - self.cds_start_i
            ce_datum = hgvs.location.CDS_START
        else:
            ce = r_interval.end.base - self.cds_end_i
            ce_datum = hgvs.location.CDS_END

        c_interval = hgvs.location.Interval(
            start=hgvs.location.BaseOffsetPosition(base=cs, offset=r_interval.start.offset, datum=cs_datum),
            end  =hgvs.location.BaseOffsetPosition(base=ce, offset=r_interval.end.offset,   datum=ce_datum),
            uncertain=r_interval.uncertain)
        return c_interval


    def c_to_r(self, c_interval):
        """convert a transcript CDS (c.) interval to a transcript cDNA (r.) interval"""

        # start
        if c_interval.start.datum == hgvs.location.CDS_START and c_interval.start.base < 0:
            rs = c_interval.start.base + self.cds_start_i + 1
        elif c_interval.start.datum == hgvs.location.CDS_START and c_interval.start.base > 0:
            rs = c_interval.start.base + self.cds_start_i
        elif c_interval.start.datum == hgvs.location.CDS_END:
            rs = c_interval.start.base + self.cds_end_i
        # end
        if c_interval.end.datum == hgvs.location.CDS_START and c_interval.end.base < 0:
            re = c_interval.end.base + self.cds_start_i + 1
        elif c_interval.end.datum == hgvs.location.CDS_START and c_interval.end.base > 0:
            re = c_interval.end.base + self.cds_start_i
        elif c_interval.end.datum == hgvs.location.CDS_END:
            re = c_interval.end.base + self.cds_end_i

        if rs <= 0:
            raise HGVSError("Transcript out of bounds.  Start position ({rs}) is <= 0.".format(rs=rs))
        if re > self.tgt_len:
            raise HGVSError("Transcript out of bounds. End position ({re}) is > than transcript length ({len}).".
                            format(re=re, len=self.tgt_len))

        r_interval = hgvs.location.Interval(
            start=hgvs.location.BaseOffsetPosition(base=rs, offset=c_interval.start.offset, datum=hgvs.location.SEQ_START),
            end  =hgvs.location.BaseOffsetPosition(base=re, offset=c_interval.end.offset, datum=hgvs.location.SEQ_START),
            uncertain=c_interval.uncertain)
        return r_interval

    def g_to_c(self, g_interval):
        """convert a genomic (g.) interval to a transcript CDS (c.) interval"""
        return self.r_to_c(self.g_to_r(g_interval))

    def c_to_g(self, c_interval):
        """convert a transcript CDS (c.) interval to a genomic (g.) interval"""
        return self.r_to_g(self.c_to_r(c_interval))



    ############################################################################
    ## DEPRECATED METHODS
    @deprecated(use_instead='g_to_c(...)')
    def hgvsg_to_hgvsc(self,*args,**kwargs):
        return self.g_to_c(*args,**kwargs)
    @deprecated(use_instead='g_to_r(...)')
    def hgvsg_to_hgvsr(self,*args,**kwargs):
        return self.g_to_r(*args,**kwargs)
    @deprecated(use_instead='r_to_g(...)')
    def hgvsr_to_hgvsg(self,*args,**kwargs):
        return self.r_to_g(*args,**kwargs)
    @deprecated(use_instead='c_to_g(...)')
    def hgvsc_to_hgvsg(self,*args,**kwargs):
        return self.c_to_g(*args,**kwargs)
    @deprecated(use_instead='c_to_r(...)')
    def hgvsc_to_hgvsr(self,*args,**kwargs):
        return self.c_to_r(*args,**kwargs)
    @deprecated(use_instead='r_to_c(...)')
    def hgvsr_to_hgvsc(self,*args,**kwargs):
        return self.r_to_c(*args,**kwargs)
    @deprecated(use_instead='c_to_p(...)')
    def hgvsc_to_hgvsp(self,*args,**kwargs):
        return self.c_to_p(*args,**kwargs)


def _strand_pm(s):
    return (None if s is None
            else '+' if s == 1
            else '-' if s == -1
            else '?')


def _hgvs_offset(g_position, grs, gre):
    """ Calculates the HGVS coordinate offset from a given genomic position"""
    if g_position == grs or g_position == gre:
        return 0
    mid = math.ceil((grs + gre) / 2)
    if g_position < mid:
        offset = g_position - grs
    else:
        offset = g_position - gre
    return offset


def _ci_to_hgvs_coord(s, e):
    """ Convert continuous interbase (right-open) coordinates (..,-2,-1,0,1,..) to
    discontinuous HGVS coordinates (..,-2,-1,1,2,..)
    """
    def _ci_to_hgvs(c):
        return c + 1 if c >= 0 else c
    return (None if s is None else _ci_to_hgvs(s),
            None if e is None else _ci_to_hgvs(e) - 1)


def _hgvs_coord_to_ci(s, e):
    """convert start,end interval in inclusive, discontinuous HGVS coordinates
    (..,-2,-1,1,2,..) to continuous interbase (right-open) coordinates
    (..,-2,-1,0,1,..)"""
    def _hgvs_to_ci(c):
        assert c != 0, 'received CDS coordinate 0; expected ..,-2,-1,1,1,...'
        return c-1 if c>0 else c
    return (None if s is None else _hgvs_to_ci(s),
            None if e is None else _hgvs_to_ci(e) + 1)


def _build_tx_cigar(exons, strand):
    cigarelem_re = re.compile('\d+[=DIMNX]')
    def _reverse_cigar(c):
        return ''.join(reversed(cigarelem_re.findall(c)))

    if len(exons) == 0:
        return None

    if strand == -1:
        for i in range(len(exons)):
            exons[i]['cigar'] = _reverse_cigar(exons[i]['cigar'])

    tx_cigar = [exons[0]['cigar']]    # exon 1
    for i in range(1, len(exons)):    # and intron + exon pairs thereafter
        tx_cigar += [str(exons[i]['alt_start_i'] - exons[i-1]['alt_end_i']) + 'N',
                     exons[i]['cigar']]
    
    tx_cigar_str = ''.join(tx_cigar)
    
    return tx_cigar_str



## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



from pkg_resources import resource_filename

import parsley

import hgvs.edit
import hgvs.hgvsposition
import hgvs.location
import hgvs.posedit
import hgvs.variant
import hgvs.utils

class Parser(object):
    """Provides comprehensive parsing of HGVS varaint strings (*i.e.*,
    variants represented according to the Human Genome Variation
    Society recommendations) into Python representations.  The class
    wraps a Parsing Expression Grammar, exposing rules of that grammar
    as methods (prefixed with `parse_`) that parse an input string
    according to the rule.  The class exposes all rules, so that it's
    possible to parse both full variant representations as well as
    components, like so:
    
    >>> hp = Parser()
    >>> v = hp.parse_hgvs_variant("NM_01234.5:c.22+1A>T")
    >>> v
    SequenceVariant(ac=NM_01234.5, type=c, posedit=22+1A>T)
    >>> v.posedit.pos
    Interval(start=22+1, end=22+1, uncertain=False)
    >>> i = hp.parse_c_interval('22+1')
    >>> i
    Interval(start=22+1, end=22+1, uncertain=False)

    The `parse_hgvs_variant` and `parse_c_interval` methods correspond
    to the `hgvs_variant` and `c_interval rules` in the grammar,
    respectively.

    Because the methods are generated on-the-fly and depend on the
    grammar that is loaded at runtime, a full list of methods is not
    available in the documentation.  However, the list of
    rules/methods is available via the `rules` instance variable.

    A few notable methods are listed below:

    `parse_hgvs_variant()` parses any valid HGVS string supported by the grammar.

      >>> hp.parse_hgvs_variant("NM_01234.5:c.22+1A>T")
      SequenceVariant(ac=NM_01234.5, type=c, posedit=22+1A>T)
      >>> hp.parse_hgvs_variant('NP_012345.6:p.Ala22Trp')
      SequenceVariant(ac=NP_012345.6, type=p, posedit=Ala22Trp)

    The `hgvs_variant` rule iteratively attempts parsing using the
    major classes of HGVS variants. For slight improvements in
    efficiency, those rules may be invoked directly:

      >>> hp.parse_p_variant('NP_012345.6:p.Ala22Trp')
      SequenceVariant(ac=NP_012345.6, type=p, posedit=Ala22Trp)

    Similarly, components of the underlying structure may be parsed
    directly as well:

      >>> hp.parse_c_posedit('22+1A>T')
      PosEdit(pos=22+1, edit=A>T, uncertain=False)
      >>> hp.parse_c_interval('22+1')
      Interval(start=22+1, end=22+1, uncertain=False)

    """

    __default_grammar_fn = resource_filename(__name__, '_data/hgvs.pymeta')

    def __init__(self,grammar_fn=__default_grammar_fn):
        self._grammar_fn = grammar_fn
        self._grammar = parsley.makeGrammar(open(grammar_fn,'r').read(),{'hgvs': hgvs})

        # define function attributes for each grammar rule, prefixed with 'parse_'
        # e.g., Parser.parse_c_interval('26+2_57-3') -> Interval(...)
        #TODO: exclude built-in rules
        self.rules = [ m.replace('rule_','')
                       for m in dir(self._grammar._grammarClass)
                       if m.startswith('rule_') ]
        for rule_name in self.rules:
            att_name = 'parse_' + rule_name
            rule_fxn = self.__make_parse_rule_function(rule_name)
            self.__setattr__(att_name,rule_fxn)

    def __make_parse_rule_function(self,rule_name):
        # http://docs.python.org/2/reference/datamodel.html#object.__getattr__
        """
        This function returns a function that takes a string and returns the parsing result.
        """
        rule_fxn = lambda s: self._grammar(s).__getattr__(rule_name)()
        rule_fxn.func_doc = "parse string s using `%s' rule" % rule_name
        return rule_fxn


## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



import pkg_resources
import warnings

try:
    __version__ = pkg_resources.get_distribution(__package__).version
except pkg_resources.DistributionNotFound as e:
    warnings.warn("can't get __version__ because %s package isn't installed" % __package__,Warning)
    __version__ = None

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



import warnings

NC_to_chr_dict = {
    'NC_000001.10':  '1', 'NC_000002.11':  '2', 'NC_000003.11': '3',
    'NC_000004.11':  '4', 'NC_000005.9' :  '5', 'NC_000006.11': '6',
    'NC_000007.13':  '7', 'NC_000008.10':  '8', 'NC_000009.11': '9',
    'NC_000010.10': '10', 'NC_000011.9' : '11', 'NC_000012.11': '12',
    'NC_000013.10': '13', 'NC_000014.8' : '14', 'NC_000015.9' : '15',
    'NC_000016.9' : '16', 'NC_000017.10': '17', 'NC_000018.9' : '18',
    'NC_000019.9' : '19', 'NC_000020.10': '20', 'NC_000021.8' : '21',
    'NC_000022.10': '22', 'NC_000023.10':  'X', 'NC_000024.9' : 'Y',
    }

chr_to_NC_dict = dict([ (v,k) for k,v in NC_to_chr_dict.iteritems() ])

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



"""
Mapping intervals between pairs of congruent segments

The IntervalMapper class is at the heart of mapping between aligned sequences.  An instance
of :class:`uta.tools.intervalmapper.IntervalMapper` is constructed with an ordered list of
:class:`uta.tools.intervalmapper.IntervalPair` instances, each of which consists of two
:class:`uta.tools.intervalmapper.Interval` instances.  The IntervalMapper class is unaware
of strand/orientation; that issue is handled by the
:class:`uta.tools.transcriptmapper.TranscriptMapper` class.

NOTE: Mapping at the boundaries around indels requires a choice.  If seq B
has an insertion relative to seq A, then mapping coordinate at the
boundaries can either be minimal or maximal for both the start and
end. Consider this alignment::

        0         15   20         35         50
        |==========|====|==========|==========|
        |          | __/        __/|          |
        |          |/          /   |          |
        |==========|==========|====|==========|
        0         15         30   35         50
            15M   5D   15M      5I      15M  

  segment 1: [ 0,15] ~ [ 0,15]
  segment 2: [15,20] ~ [15,15]
  segment 3: [20,35] ~ [15,30]
  segment 4: [35,35] ~ [30,35]
  segment 5: [35,50] ~ [35,50]

and these intervals around reference position 35::

  interval 1: 34,36   -> 29,36 (no ambiguity)
  interval 2: 35,35   -> 30,35 (reasonable)
  interval 3: 34,35   -> 29,30 (minimal) or 29,35 (maximal)
  interval 4: 35,36   -> 35,36 (minimal) or 30,36 (maximal)

So, for interval 3, end_i=35 matches segment 3 and segment 4.  Analagously
for interval 4, start_i=35 matches segment 4 and segment 5.

Currently, this code matches an interval <start_i,end_i> using the maximal
start_i and minimal end_i.
"""



import re
from hgvs.exceptions import *

class Interval(object):
    """Represents a segment of a sequence in interbase
    coordinates (0-based, right-open).
    """
    __slots__ = ('start_i','end_i')
    def __init__(self,start_i,end_i):
        if not (start_i <= end_i):
            raise HGVSInvalidIntervalError('start_i must be less than or equal to end_i')
        self.start_i = start_i
        self.end_i = end_i

    @property
    def len(self):
        return self.end_i - self.start_i

    def __repr__(self):
        return '{self.__class__.__name__}(start_i={self.start_i},end_i={self.end_i})'.format(self=self)


class IntervalPair(object):
    """Represents a match, insertion, or deletion segment of an
    alignment. If a match, the lengths must be equal; if an insertion or
    deletion, the length of the ref or tgt must be zero respectively."""
    __slots__ = ('ref','tgt')
    def __init__(self,ref,tgt):
        if not (    (ref.len == tgt.len)
                 or (ref.len == 0 and tgt.len != 0)
                 or (ref.len != 0 and tgt.len == 0) ):
            raise HGVSInvalidIntervalError("IntervalPair doesn't represent a match, insertion, or deletion")
        self.ref = ref
        self.tgt = tgt

    def __repr__(self):
        return '{self.__class__.__name__}(ref={self.ref},tgt={self.tgt})'.format(self=self)


class IntervalMapper(object):
    """Provides mapping between sequence coordinates according to an
    ordered set of IntervalPairs."""
    __slots__ = ('interval_pairs','ref_intervals','tgt_intervals','ref_len','tgt_len')
    def __init__(self,interval_pairs):
        """
        :param interval_pairs: an ordered list of IntervalPair instances
        :type interval_pairs: list (of IntervalPair instances).
        :returns: an IntervalMapper instance
        """
        def _validate_intervals(ivs):
            for i in range(1,len(ivs)):
                # check for adjacency/non-overlap
                # This constraint, combined with the start_i <= end_i constraint
                # of Intervals, guarantees that intervals are ordered
                assert ivs[i-1].end_i == ivs[i].start_i, 'intervals must be adjacent'
                
        self.interval_pairs = interval_pairs
        self.ref_intervals = [ ip.ref for ip in self.interval_pairs ]
        self.tgt_intervals = [ ip.tgt for ip in self.interval_pairs ]
        _validate_intervals(self.ref_intervals)
        _validate_intervals(self.tgt_intervals)
        self.ref_len = sum([ iv.len for iv in self.ref_intervals ])
        self.tgt_len = sum([ iv.len for iv in self.tgt_intervals ])
        

    @staticmethod
    def from_cigar(cigar):
        """
        :param cigar: a Compact Idiosyncratic Gapped Alignment Report string
        :type cigar: str.
        :returns: an IntervalMapper instance from the CIGAR string
        """
        return IntervalMapper(cigar_to_intervalpairs(cigar))

    def map_ref_to_tgt(self,start_i,end_i,max_extent=False):
        return self._map(self.ref_intervals, self.tgt_intervals, start_i, end_i, max_extent)

    def map_tgt_to_ref(self,start_i,end_i,max_extent=False):
        return self._map(self.tgt_intervals, self.ref_intervals, start_i, end_i, max_extent)

    @staticmethod
    def _map(from_ivs,to_ivs,from_start_i,from_end_i,max_extent):
        def iv_map(from_ivs,to_ivs,from_start_i,from_end_i,max_extent):
            """returns the <start,end> intervals indexes in which from_start_i and from_end_i occur"""
            # first look for 0-width interval that matches
            seil = [ i for i,iv in enumerate(from_ivs) if iv.start_i == from_start_i and iv.end_i == from_end_i ]
            if len(seil) > 0:
                si = ei = seil[0]
            else:
                sil = [ i for i,iv in enumerate(from_ivs) if iv.start_i <= from_start_i <= iv.end_i ]
                eil = [ i for i,iv in enumerate(from_ivs) if iv.start_i <= from_end_i   <= iv.end_i ]
                if len(sil) == 0 or len(eil) == 0:
                    raise HGVSInvalidIntervalError('start or end or both are beyond the bounds of transcript record')
                si,ei = (sil[0],eil[-1]) if max_extent else (sil[-1],eil[0])
            return si,ei

        def clip_to_iv(iv,pos):
            return max(iv.start_i,min(iv.end_i,pos))

        assert from_start_i<=from_end_i, 'expected from_start_i <= from_end_i'
        try:
            si,ei = iv_map(from_ivs,to_ivs,from_start_i,from_end_i,max_extent)
        except ValueError:
            raise HGVSInvalidIntervalError('start_i,end_i interval out of bounds')
        to_start_i = clip_to_iv( to_ivs[si], to_ivs[si].start_i + (from_start_i - from_ivs[si].start_i) )
        to_end_i   = clip_to_iv( to_ivs[ei], to_ivs[ei].end_i   - (from_ivs[ei].end_i - from_end_i)     )
        return to_start_i,to_end_i


class CIGARElement(object):
    """represents elements of a CIGAR string and provides methods for
    determining the number of ref and tgt bases consumed by the
    operation"""

    __slots__ = ('len','op')
    def __init__(self,len,op):
        self.len = len
        self.op = op

    @property
    def ref_len(self):
        """returns number of nt/aa consumed in reference sequence for this edit"""
        return self.len if self.op in '=INX' else 0

    @property
    def tgt_len(self):
        """returns number of nt/aa consumed in target sequence for this edit"""
        return self.len if self.op in '=DX'  else 0


def cigar_to_intervalpairs(cigar):
    """For a given CIGAR string, return a list of (Interval,Interval)
    pairs.  The length of the returned list will be equal to the
    number of CIGAR operations
    """

    cigar_elem_re = re.compile('(?P<len>\d+)(?P<op>[=DIMNX])')
    ces = [ CIGARElement(op=md['op'],len=int(md['len']))
            for md in [ m.groupdict() for m in cigar_elem_re.finditer(cigar) ] ]
    ips = [None] * len(ces)
    ref_pos = tgt_pos = 0
    for i,ce in enumerate(ces):
        ips[i] = IntervalPair( Interval(ref_pos,ref_pos+ce.ref_len),
                               Interval(tgt_pos,tgt_pos+ce.tgt_len) )
        ref_pos += ce.ref_len
        tgt_pos += ce.tgt_len
    return ips



## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



import copy

from Bio.Seq import Seq
import recordtype

import hgvs.exceptions
import hgvs.location
import hgvs.posedit
import hgvs.transcriptmapper
import hgvs.utils.altseq_to_hgvsp as altseq_to_hgvsp
import hgvs.utils.altseqbuilder as altseqbuilder
import hgvs.variant

from hgvs.utils import reverse_complement
from hgvs.utils.deprecated import deprecated



class VariantMapper(object):
    """
    Maps HGVS variants to and from g., r., c., and p. representations.
    All methods require and return objects of type :class:`hgvs.variant.SequenceVariant`.
    """

    def __init__(self,hdp,cache_transcripts=False):
        self.hdp = hdp
        self.cache_transcripts = cache_transcripts
        self.__tm_cache = {}


    def g_to_c(self, var_g, tx_ac, alt_aln_method='splign'):
        """Given a genomic (g.) parsed HGVS variant, return a transcript (c.) variant on the specified transcript using
        the specified alignment method (default is 'splign' from NCBI).

        :param hgvs.variant.SequenceVariant var_g: a variant object
        :param str tx_ac: a transcript accession (e.g., NM_012345.6 or ENST012345678)
        :param str alt_aln_method: the alignment method; valid values depend on data source
        :returns: variant object (:class:`hgvs.variant.SequenceVariant`) using CDS coordinates
        :raises hgvs.exceptions.HGVSInvalidVariantError: if var_g is not of type 'g'

        """

        if not (var_g.type == 'g'):
            raise hgvs.exceptions.HGVSInvalidVariantError('Expected a genomic (g.) variant; got '+ str(var_g))

        tm = self._fetch_TranscriptMapper(tx_ac=tx_ac, alt_ac=var_g.ac, alt_aln_method=alt_aln_method)
        
        pos_c = tm.g_to_c( var_g.posedit.pos )
        edit_c = self._convert_edit_check_strand(tm.strand, var_g.posedit.edit)
        var_c = hgvs.variant.SequenceVariant(ac=tx_ac,
                                             type='c',
                                             posedit=hgvs.posedit.PosEdit( pos_c, edit_c ) )
        return var_c


    def g_to_r(self, var_g, tx_ac, alt_aln_method='splign'):
        """Given a genomic (g.) parsed HGVS variant, return a transcript (r.) variant on the specified transcript using
        the specified alignment method (default is 'splign' from NCBI).

        :param hgvs.variant.SequenceVariant var_g: a variant object
        :param str tx_ac: a transcript accession (e.g., NM_012345.6 or ENST012345678)
        :param str alt_aln_method: the alignment method; valid values depend on data source
        :returns: variant object (:class:`hgvs.variant.SequenceVariant`) using transcript (r.) coordinates
        :raises hgvs.exceptions.HGVSInvalidVariantError: if var_g is not of type 'g'

        """

        if not (var_g.type == 'g'):
            raise hgvs.exceptions.HGVSInvalidVariantError('Expected a genomic (g.); got '+ str(var_g))

        tm = self._fetch_TranscriptMapper(tx_ac=tx_ac, alt_ac=var_g.ac, alt_aln_method=alt_aln_method)

        pos_r = tm.g_to_r( var_g.posedit.pos )
        edit_r = self._convert_edit_check_strand(tm.strand, var_g.posedit.edit)
        var_r = hgvs.variant.SequenceVariant(ac=tx_ac,
                                             type='r',
                                             posedit=hgvs.posedit.PosEdit( pos_r, edit_r ) )
        return var_r


    def r_to_g(self, var_r, alt_ac, alt_aln_method='splign'):
        """Given an RNA (r.) parsed HGVS variant, return a genomic (g.) variant on the specified transcript using
        the specified alignment method (default is 'splign' from NCBI).

        :param hgvs.variant.SequenceVariant var_r: a variant object
        :param str alt_ac: a reference sequence accession (e.g., NC_000001.11)
        :param str alt_aln_method: the alignment method; valid values depend on data source
        :returns: variant object (:class:`hgvs.variant.SequenceVariant`)
        :raises hgvs.exceptions.HGVSInvalidVariantError: if var_r is not of type 'r'

        """

        if not (var_r.type == 'r'):
            raise hgvs.exceptions.HGVSInvalidVariantError('Expected a RNA (r.); got '+ str(var_r))

        tm = self._fetch_TranscriptMapper(tx_ac=var_r.ac, alt_ac=alt_ac, alt_aln_method=alt_aln_method)

        pos_g = tm.r_to_g( var_r.posedit.pos )
        edit_g = self._convert_edit_check_strand(tm.strand, var_r.posedit.edit)

        var_g = hgvs.variant.SequenceVariant(ac=alt_ac,
                                             type='g',
                                             posedit=hgvs.posedit.PosEdit( pos_g, edit_g ) )
        return var_g

    
    def c_to_g(self, var_c, alt_ac, alt_aln_method='splign'):
        """Given a cDNA (c.) parsed HGVS variant, return a genomic (g.) variant on the specified transcript using
        the specified alignment method (default is 'splign' from NCBI).

        :param hgvs.variant.SequenceVariant var_c: a variant object
        :param str alt_ac: a reference sequence accession (e.g., NC_000001.11)
        :param str alt_aln_method: the alignment method; valid values depend on data source
        :returns: variant object (:class:`hgvs.variant.SequenceVariant`)
        :raises hgvs.exceptions.HGVSInvalidVariantError: if var_c is not of type 'c'

        """

        if not (var_c.type == 'c'):
            raise hgvs.exceptions.HGVSInvalidVariantError('Expected a cDNA (c.); got ' + str(var_c))

        tm = self._fetch_TranscriptMapper(tx_ac=var_c.ac, alt_ac=alt_ac, alt_aln_method=alt_aln_method)

        pos_g = tm.c_to_g(var_c.posedit.pos)
        edit_g = self._convert_edit_check_strand(tm.strand, var_c.posedit.edit)

        var_g = hgvs.variant.SequenceVariant(ac=alt_ac,
                                             type='g',
                                             posedit=hgvs.posedit.PosEdit(pos_g, edit_g))
        return var_g


    def c_to_r(self, var_c):
        """Given a cDNA (c.) parsed HGVS variant, return a RNA (r.) variant on the specified transcript using
        the specified alignment method (default is 'transcript' indicating a self alignment).

        :param hgvs.variant.SequenceVariant var_c: a variant object
        :returns: variant object (:class:`hgvs.variant.SequenceVariant`)
        :raises hgvs.exceptions.HGVSInvalidVariantError: if var_c is not of type 'c'

        """

        if not (var_c.type == 'c'):
            raise hgvs.exceptions.HGVSInvalidVariantError('Expected a cDNA (c.); got ' + str(var_c))

        tm = self._fetch_TranscriptMapper(tx_ac=var_c.ac, alt_ac=var_c.ac, alt_aln_method='transcript')
        pos_r = tm.c_to_r(var_c.posedit.pos)

        # not necessary to check strand
        if isinstance(var_c.posedit.edit, hgvs.edit.NARefAlt) or isinstance(var_c.posedit.edit, hgvs.edit.Dup):
            edit_r = var_c.posedit.edit
        else:
            raise NotImplemented('Only NARefAlt/Dup types are currently implemented')

        var_r = hgvs.variant.SequenceVariant(ac=var_c.ac,
                                             type='r',
                                             posedit=hgvs.posedit.PosEdit( pos_r, edit_r ) )
        return var_r


    def r_to_c(self, var_r):
        """Given an RNA (r.) parsed HGVS variant, return a cDNA (c.) variant on the specified transcript using
        the specified alignment method (default is 'transcript' indicating a self alignment).

        :param hgvs.variant.SequenceVariant var_r: a variant object
        :returns: variant object (:class:`hgvs.variant.SequenceVariant`)
        :raises hgvs.exceptions.HGVSInvalidVariantError: if var_r is not of type 'r'

        """

        if not (var_r.type == 'r'):
            raise hgvs.exceptions.HGVSInvalidVariantError('Expected RNA (r.); got ' + str(var_r))

        tm = self._fetch_TranscriptMapper(tx_ac=var_r.ac, alt_ac=var_r.ac, alt_aln_method='transcript')
        pos_c = tm.r_to_c(var_r.posedit.pos)

        # not necessary to check strand
        if isinstance(var_r.posedit.edit, hgvs.edit.NARefAlt) or isinstance(var_r.posedit.edit, hgvs.edit.Dup):
            edit_c = var_r.posedit.edit
        else:
            raise NotImplemented('Only NARefAlt types are currently implemented')

        var_c = hgvs.variant.SequenceVariant(ac=var_r.ac,
                                             type='c',
                                             posedit=hgvs.posedit.PosEdit( pos_c, edit_c ) )
        return var_c


    # TODO: c_to_p needs refactoring
    # TODO: data prep belongs in the data interface
    def c_to_p(self, var_c, pro_ac=None):
        """
        Converts a c. SequenceVariant to a p. SequenceVariant on the specified protein accession
        Author: Rudy Rico

        :param SequenceVariant var_c: hgvsc tag
        :param str pro_ac: protein accession
        :rtype: hgvs.variant.SequenceVariant

        """

        class RefTranscriptData(recordtype.recordtype('RefTranscriptData',
                                                      ['transcript_sequence', 'aa_sequence',
                                                       'cds_start', 'cds_stop', 'protein_accession'])):

            @classmethod
            def setup_transcript_data(cls, hdp, tx_ac, pro_ac):
                """helper for generating RefTranscriptData from for c_to_p"""
                tx_info = hdp.get_tx_identity_info(var_c.ac)
                tx_seq = hdp.get_tx_seq(tx_ac)

                if tx_info is None or tx_seq is None:
                    raise hgvs.exceptions.HGVSError("Missing transcript data for accession: {}".format(tx_ac))

                # use 1-based hgvs coords
                cds_start = tx_info['cds_start_i'] + 1
                cds_stop = tx_info['cds_end_i']

                # padding list so biopython won't complain during the conversion
                tx_seq_to_translate = tx_seq[cds_start - 1:cds_stop]
                if len(tx_seq_to_translate) % 3 != 0:
                    ''.join(list(tx_seq_to_translate).extend(['N']*((3-len(tx_seq_to_translate) % 3) % 3)))

                tx_seq_cds = Seq(tx_seq_to_translate)
                protein_seq = str(tx_seq_cds.translate())

                if pro_ac is None:
                    # get_acs... will always return at least the MD5_ accession
                    pro_ac = hdp.get_acs_for_protein_seq(protein_seq)[0]

                transcript_data = RefTranscriptData(tx_seq, protein_seq, cds_start,
                                                    cds_stop, pro_ac)

                return transcript_data

        if not (var_c.type == 'c'):
            raise hgvs.exceptions.HGVSInvalidVariantError('Expected a cDNA (c.); got ' + str(var_c))

        reference_data = RefTranscriptData.setup_transcript_data(self.hdp, var_c.ac, pro_ac)
        builder = altseqbuilder.AltSeqBuilder(var_c, reference_data)

        # TODO - handle case where you get 2+ alt sequences back; currently get list of 1 element
        # loop structure implemented to handle this, but doesn't really do anything currently.
        all_alt_data = builder.build_altseq()

        var_ps = []
        for alt_data in all_alt_data:
            builder = altseq_to_hgvsp.AltSeqToHgvsp(reference_data, alt_data)
            var_p = builder.build_hgvsp()
            var_ps.append(var_p)

        var_p = var_ps[0]

        return var_p


    ############################################################################
    ## DEPRECATED METHODS
    @deprecated(use_instead='g_to_c(...)')
    def hgvsg_to_hgvsc(self,*args,**kwargs):
        return self.g_to_c(*args,**kwargs)
    @deprecated(use_instead='g_to_r(...)')
    def hgvsg_to_hgvsr(self,*args,**kwargs):
        return self.g_to_r(*args,**kwargs)
    @deprecated(use_instead='r_to_g(...)')
    def hgvsr_to_hgvsg(self,*args,**kwargs):
        return self.r_to_g(*args,**kwargs)
    @deprecated(use_instead='c_to_g(...)')
    def hgvsc_to_hgvsg(self,*args,**kwargs):
        return self.c_to_g(*args,**kwargs)
    @deprecated(use_instead='c_to_r(...)')
    def hgvsc_to_hgvsr(self,*args,**kwargs):
        return self.c_to_r(*args,**kwargs)
    @deprecated(use_instead='r_to_c(...)')
    def hgvsr_to_hgvsc(self,*args,**kwargs):
        return self.r_to_c(*args,**kwargs)
    @deprecated(use_instead='c_to_p(...)')
    def hgvsc_to_hgvsp(self,*args,**kwargs):
        return self.c_to_p(*args,**kwargs)


    ############################################################################
    ## Internal methods

    def _fetch_TranscriptMapper(self, tx_ac, alt_ac, alt_aln_method):
        """
        Get a new TranscriptMapper for the given transcript accession (ac),
        possibly caching the result.
        """
        key = '_'.join([tx_ac, alt_ac, alt_aln_method])
        try:
            tm = self.__tm_cache[key]
        except KeyError:
            tm = hgvs.transcriptmapper.TranscriptMapper(self.hdp, tx_ac=tx_ac, alt_ac=alt_ac,
                                                        alt_aln_method=alt_aln_method)
            if self.cache_transcripts:
                self.__tm_cache[key] = tm
        return tm


    @staticmethod
    def _convert_edit_check_strand(strand, edit_in):
        """
        Convert an edit from one type to another, based on the stand and type
        """
        if isinstance(edit_in, hgvs.edit.NARefAlt):
            if strand == 1:
                edit_out = copy.deepcopy(edit_in)
            else:
                try:
                    # if smells like an int, do nothing
                    # TODO: should use ref_n, right?
                    int(edit_in.ref)
                    ref = edit_in.ref
                except (ValueError, TypeError):
                    ref = reverse_complement(edit_in.ref)
                edit_out = hgvs.edit.NARefAlt(
                    ref = ref,
                    alt = reverse_complement(edit_in.alt),
                    )
        elif isinstance(edit_in, hgvs.edit.Dup):
            if strand == 1:
                edit_out = copy.deepcopy(edit_in)
            else:
                edit_out = hgvs.edit.Dup(
                    seq = reverse_complement(edit_in.seq)
                )
        else:
            raise NotImplemented('Only NARefAlt/Dup types are currently implemented')
        return edit_out


# TODO: move assembly data to UTA and query through HDPI
primary_assembly_accessions = {
    'GRCh37': {
        'NC_000001.10', 'NC_000002.11', 'NC_000003.11',
        'NC_000004.11', 'NC_000005.9' , 'NC_000006.11',
        'NC_000007.13', 'NC_000008.10', 'NC_000009.11',
        'NC_000010.10', 'NC_000011.9' , 'NC_000012.11',
        'NC_000013.10', 'NC_000014.8',  'NC_000015.9',
        'NC_000016.9' , 'NC_000017.10', 'NC_000018.9',
        'NC_000019.9' , 'NC_000020.10', 'NC_000021.8',
        'NC_000022.10', 'NC_000023.10', 'NC_000024.9',
        },
    }

class EasyVariantMapper(VariantMapper):
    """Provides simplified variant mapping for a single assembly and
    transcript-reference alignment method.
    
    EasyVariantMapper is instantiated with a primary_assembly and
    alt_aln_method. These enable the following conveniences over
    VariantMapper:

    * The primary assembly and alignment method are used to
      automatically select an appropriate chromosomal reference
      sequence when mapping from a transcript to a genome (i.e.,
      c_to_g(...) and r_to_g(...)).

    * A new method, relevant_trancripts(g_variant), returns a list of
      transcript accessions available for the specified variant. These
      accessions are candidates mapping from genomic to trancript
      coordinates (i.e., g_to_c(...) and g_to_r(...)).

    [tests occur in module doc (rather than in method doc) to use a
    single db connection]

    IMPORTANT: Callers should be prepared to catch HGVSError
    exceptions. These will be thrown whenever a transcript maps
    ambiguously to a chromosome, such as for pseudoautosomal region
    transcripts.
    """

    def __init__(self,hdp,primary_assembly='GRCh37',alt_aln_method='splign',cache_transcripts=False):
        super(EasyVariantMapper,self).__init__(hdp=hdp,cache_transcripts=cache_transcripts)
        self.primary_assembly = primary_assembly
        self.alt_aln_method = alt_aln_method
        self.primary_assembly_accessions = set(primary_assembly_accessions[primary_assembly])

    def g_to_c(self, var_g, tx_ac): 
        return super(EasyVariantMapper,self).g_to_c(var_g, tx_ac, alt_aln_method=self.alt_aln_method)
    def g_to_r(self, var_g, tx_ac): 
        return super(EasyVariantMapper,self).g_to_r(var_g, tx_ac, alt_aln_method=self.alt_aln_method)

    def c_to_g(self, var_c): 
        alt_ac = self._alt_ac_for_tx_ac(var_c.ac)
        return super(EasyVariantMapper,self).c_to_g(var_c, alt_ac, alt_aln_method=self.alt_aln_method)
    def r_to_g(self, var_r): 
        alt_ac = self._alt_ac_for_tx_ac(var_r.ac)
        return super(EasyVariantMapper,self).r_to_g(var_r, alt_ac, alt_aln_method=self.alt_aln_method)

    def c_to_r(self, var_c): 
        return super(EasyVariantMapper,self).c_to_r(var_c)
    def r_to_c(self, var_r): 
        return super(EasyVariantMapper,self).r_to_c(var_r)

    def c_to_p(self, var_c): 
        return super(EasyVariantMapper,self).c_to_p(var_c)

    def relevant_transcripts(self,var_g):
        """return list of transcripts accessions (strings) for given variant,
        selected by genomic overlap"""
        tx = self.hdp.get_tx_for_region(var_g.ac,
                                        self.alt_aln_method,
                                        var_g.posedit.pos.start.base,
                                        var_g.posedit.pos.end.base)
        return [ e['tx_ac'] for e in tx ]

    def _alt_ac_for_tx_ac(self,tx_ac):
        """return chromosomal accession for given transcript accession (and
        the primary_assembly and aln_method setting used to
        instantiate this EasyVariantMapper)

        """
        alt_acs = [e['alt_ac'] 
                   for e in  self.hdp.get_tx_mapping_options(tx_ac)
                   if e['alt_aln_method'] == self.alt_aln_method
                   and e['alt_ac'] in self.primary_assembly_accessions]
        if len(alt_acs) > 1:
            raise hgvs.exceptions.HGVSError("Multiple chromosomal alignments for {tx_ac} in {pa} using {am} (likely paralog or pseudoautosomal region)".format(
                tx_ac=tx_ac, pa=self.primary_assembly, am=self.alt_aln_method))
        if len(alt_acs) == 0:
            raise hgvs.exceptions.HGVSError("No alignments for {tx_ac} in {pa} using {am}".format(
                tx_ac=tx_ac, pa=self.primary_assembly, am=self.alt_aln_method))
        return alt_acs[0]       # exactly one remains



## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



import base64
import hashlib
import string

aa3_to_aa1_lut = {
    'Ala': 'A',    'Arg': 'R',    'Asn': 'N',    'Asp': 'D',
    'Cys': 'C',    'Gln': 'Q',    'Glu': 'E',    'Gly': 'G',
    'His': 'H',    'Ile': 'I',    'Leu': 'L',    'Lys': 'K',
    'Met': 'M',    'Phe': 'F',    'Pro': 'P',    'Ser': 'S',
    'Thr': 'T',    'Trp': 'W',    'Tyr': 'Y',    'Val': 'V',
    'Xaa': 'X',    'Ter': '*',    'Sec': 'U',
    }

aa1_to_aa3_lut = { v: k for k,v in aa3_to_aa1_lut.iteritems() }


def aa3_to_aa1(s):
    "convert string of 3-letter amino acids to 1-letter amino acids"
    return None if s is None else  ''.join([aa3_to_aa1_lut[aa3] for aa3 in
                                            [ s[i:i+3] for i in range(0,len(s),3) ]])

def aa1_to_aa3(s):
    "convert string of 1-letter amino acids to 3-letter amino acids"
    return None if s is None else ''.join([aa1_to_aa3_lut[aa1] for aa1 in s])

def aa_to_aa1(s):
    "coerce string of 1- or 3-letter amino acids to 1-letter"
    return aa3_to_aa1(s) if __looks_like_aa3_p(s) else s

def aa_to_aa3(s):
    "coerce string of 1- or 3-letter amino acids to 3-letter"
    return aa1_to_aa3(s) if not __looks_like_aa3_p(s) else s


def seq_seguid(seq):
    """returns BioPython-compatible seguid""" 
    seq = seq.upper().rstrip('*')
    return base64.standard_b64encode(hashlib.sha1(seq).digest()).rstrip('=')

def seq_md5(seq):
    "returns sequence md5 as hex digest"
    seq = seq.upper().rstrip('*')
    return hashlib.md5(seq).hexdigest()

def seq_sha1(seq):
    """returns sequence sha1 as url-safe base64 encoded digest. This is
    identical to seguid except for hashes that contain / or +."""
    seq = seq.upper().rstrip('*')
    return base64.urlsafe_b64encode(hashlib.sha1(seq).digest()).rstrip('=')


def __looks_like_aa3_p(s):
    "string looks like a 3-letter AA string"
    return (
        s is not None
        and (len(s) % 3 == 0)
        and (len(s) == 0 or s[1] in string.lowercase)
        )

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- encoding: utf-8 -*-

"""
hgvs.location -- classes for dealing with the locations of HGVS variants

This module provides for Representing the location of variants in HGVS nomenclature, including:

  * integers and integer intervals (e.g., NC_012345.6:g.3403243_3403248A>C)
  * CDS positions and intervals (e.g., NM_01234.5:c.56+12_56+14delAC)
  * CDS stop coordinates (e.g., NM_01234.5:c.*13A>C)  

Classes:

  * :class:`SimplePosition` -- a simple integer
  * :class:`BaseOffsetPosition` -- a position with datum, base, and offset for c. and r. coordinates
  * :class:`AAPosition` -- an amino acid position (with AA)
  * :class:`Interval` -- an interval of Positions
"""


import recordtype

from hgvs.utils import aa1_to_aa3

SEQ_START = 0
CDS_START = 1
CDS_END = 2

class SimplePosition( recordtype.recordtype(
        'SimplePosition',
        field_names = [ ('base', None), ('uncertain', False) ]
        )):

    def __str__(self):
        self.validate()
        s = '?' if self.base is None else str(self.base)
        return '('+s+')' if self.uncertain else s

    @property
    def is_uncertain(self):
        """return True if the position is marked uncertain or undefined"""
        return self.uncertain or self.base in None

    def _set_uncertain(self):
        "mark this location as uncertain and return reference to self; this is called during parsing (see hgvs.ometa)"
        self.uncertain = True
        return self

    def validate(self):
        "raise AssertionError if instance variables are invalid; otherwise return True"
        assert self.base is None or self.base>=1, self.__class__.__name__ + ': base must be >= 1'
        return True


class BaseOffsetPosition( recordtype.recordtype(
        'BaseOffsetPosition',
        field_names = [ ('base',None),
                        ('offset',0),
                        ('datum',SEQ_START),
                        ('uncertain',False) ]
        )):
    """
    Class for dealing with CDS coordinates in transcript variants.

    This class models CDS positions using a `base` coordinate, which is
    measured relative to a specified `datum` (CDS_START or CDS_END), and
    an `offset`, which is 0 for exonic positions and non-zero for intronic
    positions.  **Positions and offsets are 1-based**, with no 0, per the HGVS
    recommendations.  (If you're using this with UTA, be aware that UTA
    uses interbase coordinates.)

    +----------+------------+-------+---------+------------------------------------------+
    | hgvs     | datum      | base  | offset  | meaning                                  |
    +==========+============+=======+=========+==========================================+
    | r.55     | SEQ_START  |   55  |      0  | RNA position 55                          |
    +----------+------------+-------+---------+------------------------------------------+
    | c.55     | CDS_START  |   55  |      0  | CDS position 55                          |
    +----------+------------+-------+---------+------------------------------------------+
    | c.55     | CDS_START  |   55  |      0  | CDS position 55                          |
    +----------+------------+-------+---------+------------------------------------------+
    | c.55+1   | CDS_START  |   55  |      1  | intronic variant +1 from boundary        |
    +----------+------------+-------+---------+------------------------------------------+
    | c.-55    | CDS_START  |  -55  |      0  | 5' UTR variant, 55 nt upstream of ATG    |
    +----------+------------+-------+---------+------------------------------------------+
    | c.1      | CDS_START  |    1  |      0  |   start codon                            |
    +----------+------------+-------+---------+------------------------------------------+
    | c.1234   | CDS_START  | 1234  |      0  | stop codon (assuming CDS length is 1233) |
    +----------+------------+-------+---------+------------------------------------------+
    | c.*1     | CDS_END    |    0  |      1  | STOP + 1                                 |
    +----------+------------+-------+---------+------------------------------------------+
    | c.*55    | CDS_END    |    0  |     55  | 3' UTR variant, 55 nt after STOP         |
    +----------+------------+-------+---------+------------------------------------------+
    """
    
    def validate(self):
        "raise AssertionError if instance variables are invalid; otherwise return True"
        assert self.base is None or self.base != 0, 'BaseOffsetPosition base may not be 0'
        assert self.base is None or self.datum == CDS_START or self.base >= 1, 'BaseOffsetPosition base must be >=1 for datum = SEQ_START or CDS_END'
        return True

    def __str__(self):
        self.validate()
        base_str = ( '?' if self.base is None
                     else '*' + str(self.base) if self.datum == CDS_END
                     else str(self.base) )
        offset_str = ( '+?' if self.offset is None
                       else '' if self.offset == 0
                       else '%+d' % self.offset )
        pos = base_str + offset_str
        return '('+pos+')' if self.uncertain else pos

    def _set_uncertain(self):
        "mark this location as uncertain and return reference to self; this is called during parsing (see hgvs.ometa)"
        self.uncertain = True
        return self

    @property
    def is_uncertain(self):
        """return True if the position is marked uncertain or undefined"""
        return self.uncertain or self.base is None or self.offset is None


class AAPosition( recordtype.recordtype(
        'AAPosition', field_names = [ ('base',None), ('aa',None), ('uncertain',False) ] ) ):

    def validate(self):
        "raise AssertionError if instance variables are invalid; otherwise return True"
        assert self.base is None or self.base >= 1, 'AAPosition location must be >=1'
        assert len(self.aa) == 1, 'More than 1 AA associated with position'
        return True

    def __str__(self):
        self.validate()
        pos = '?' if self.base is None else str(self.base)
        aa = '?' if self.aa is None else aa1_to_aa3(self.aa)
        s = aa + pos
        return '('+s+')' if self.uncertain else s

    @property
    def pos(self):
        """return base, for backward compatibility"""
        return self.base

    def _set_uncertain(self):
        "mark this location as uncertain and return reference to self; this is called during parsing (see hgvs.ometa)"
        self.uncertain = True
        return self

    @property
    def is_uncertain(self):
        """return True if the position is marked uncertain or undefined"""
        return self.uncertain or self.base is None or self.aa is None

class Interval( recordtype.recordtype(
        'Interval', field_names = [ 'start', ('end',None), ('uncertain',False) ] ) ):

    def validate(self):
        "raise AssertionError if instance variables are invalid; otherwise return True"
        return True

    def __str__(self):
        self.validate()
        if self.end is None or self.start == self.end:
            return str(self.start)
        iv = str(self.start) + '_' + str(self.end)
        return '('+iv+')' if self.uncertain else iv

    def _set_uncertain(self):
        "mark this interval as uncertain and return reference to self; this is called during parsing (see hgvs.ometa)"
        self.uncertain = True
        return self

    @property
    def is_uncertain(self):
        """return True if the position is marked uncertain or undefined"""
        return self.uncertain or self.start.is_uncertain or self.end.is_uncertain


## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



import sys, os

import hgvs
version = hgvs.__version__
release = str(hgvs.__version__)

project = u'HGVS'
copyright = u'2014, Invitae'

source_suffix = '.rst'
master_doc = 'index'
exclude_patterns = ['build','static','templates','themes']
pygments_style = 'sphinx'
templates_path = ['templates']

html_theme = 'invitae'
html_theme_path = ['../themes']
html_title = '{project} {release}'.format(project = project, release = release)
html_logo = '../static/hgvs-logo.svg'
html_favicon = '../static/favicon.ico'
html_static_path = ['../static']

autodoc_default_flags = ['members', 'undoc-members', 'show-inheritance'] #, 'inherited-members']

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.intersphinx',
    'sphinx.ext.pngmath',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
    'sphinxcontrib.fulltoc',
    ]

extlinks={
    'issue': ('https://bitbucket.org/invitae/hgvs/issue/%s' 'HGVS issue '),
    }

intersphinx_mapping = {
    'http://docs.python.org/': None,
    }

latex_documents = [
  ('index', 'HGVS.tex', u'HGVS Documentation', u'HGVS Contributors', 'manual'),
]

man_pages = [
    ('index', 'uta', u'HGVS Documentation', [u'HGVS Contributors'], 1)
]

texinfo_documents = [
  ('index', 'HGVS', u'HGVS Documentation',
   u'HGVS Contributors', 'HGVS', 'One line description of project.',
   'Miscellaneous'),
]



import unittest

from nose.plugins.attrib import attr

import hgvs.location

@attr(tags=["quick","models"])
class Test_SimplePosition(unittest.TestCase):
    def test_success(self):
        self.assertEqual( str(hgvs.location.SimplePosition(5)), '5' )
        self.assertEqual( str(hgvs.location.SimplePosition(5,uncertain=True)), '(5)' )
        self.assertEqual( str(hgvs.location.SimplePosition(None)), '?' )

    def test_failure(self):
        with self.assertRaises(AssertionError):
            self.assertEqual( hgvs.location.SimplePosition(-1), 'SHOULD FAIL' )


@attr(tags=["quick"])
class Test_BaseOffsetPosition(unittest.TestCase):
    def test_success(self):
        # r.5
        cdsp = hgvs.location.BaseOffsetPosition(5)
        self.assertEqual( cdsp.datum, hgvs.location.SEQ_START )
        self.assertEqual( cdsp.base, 5 )
        self.assertEqual( cdsp.offset, 0 )
        self.assertEqual( str(cdsp), '5' )

        #r.5+6
        cdsp.offset = 6
        self.assertEqual( str(cdsp), '5+6' )

        #r.5+?
        cdsp.offset = None
        self.assertEqual( str(cdsp), '5+?' )

        #r.(5+?)
        cdsp.uncertain = True
        self.assertEqual( str(cdsp), '(5+?)' )
        
        # c.*5
        cdsp = hgvs.location.BaseOffsetPosition(5,datum=hgvs.location.CDS_END)
        self.assertEqual( cdsp.datum, hgvs.location.CDS_END )
        self.assertEqual( cdsp.base, 5 )
        self.assertEqual( cdsp.offset, 0 )
        self.assertEqual( str(cdsp), '*5' )

        cdsp.uncertain = True
        self.assertEqual( str(cdsp), '(*5)' )
        
        cdsp.offset = 7
        self.assertEqual( str(cdsp), '(*5+7)' )


@attr(tags=["quick"])
class Test_AAPosition(unittest.TestCase):
    def test_AAPosition(self):
        ap = hgvs.location.AAPosition( 15, 'S' )
        self.assertEqual( ap.pos, 15 )
        self.assertEqual( str(ap), 'Ser15' )

        
@attr(tags=["quick"])
class Test_Interval(unittest.TestCase):        
    def test_Interval(self):
        ival = hgvs.location.Interval( hgvs.location.BaseOffsetPosition(base=12,offset=+34),
                                       hgvs.location.BaseOffsetPosition(base=56,offset=-78) )
        self.assertEqual( ival.start.base, 12 )
        self.assertEqual( ival.start.offset, 34 )
        self.assertEqual( ival.end.base, 56 )
        self.assertEqual( ival.end.offset, -78 )
        self.assertEqual( str(ival), '12+34_56-78' )

if __name__ == '__main__':
    unittest.main()

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



from __future__ import with_statement

import unittest

from nose.plugins.attrib import attr

import hgvs.exceptions
import hgvs.intervalmapper

@attr(tags=["quick"])
class Test_IntervalMapper(unittest.TestCase):
    longMessage = True

    def test_interval_valid(self):
        interval = hgvs.intervalmapper.Interval(3, 4)
        st = str(interval)
        self.assertEqual(1, interval.len)
        self.assertEqual(interval.__repr__(), 'Interval(start_i=3,end_i=4)')

    def test_interval_invalid_start_gt_end(self):
        with self.assertRaises(hgvs.exceptions.HGVSInvalidIntervalError):
            interval = hgvs.intervalmapper.Interval(4,3)

    def test_intervalpair_valid_match(self):
        self._check_valid_intervalpair(3, 4, 6, 7)

    def test_intervalpair_valid_insertion(self):
        self._check_valid_intervalpair(3, 3, 6, 7)

    def test_intervalpair_valid_deletion(self):
        self._check_valid_intervalpair(3, 4, 6, 6)

    def test_intervalpair_invalid(self):
        iv1 = hgvs.intervalmapper.Interval(3, 4)
        iv2 = hgvs.intervalmapper.Interval(6, 9)
        with self.assertRaises(hgvs.exceptions.HGVSInvalidIntervalError):
            intervalpair = hgvs.intervalmapper.IntervalPair(iv1, iv2)

    def test_intervalmapper_valid_ranges(self):
        ivm = self._build_mock_intervalmapper()

        self.assertEqual((9, 8), (ivm.ref_len, ivm.tgt_len))
        self.assertEqual((1, 8), ivm.map_tgt_to_ref(1, 7))
        self.assertEqual((1, 6), ivm.map_ref_to_tgt(1, 7))

    def test_intervalmapper_invalid_start_gt_end(self):
        ivm = self._build_mock_intervalmapper()
        with self.assertRaises(AssertionError):
            (s, e) = ivm.map_tgt_to_ref(5, 4)

    def test_intervalmapper_invalid_out_of_range(self):
        ivm = self._build_mock_intervalmapper()
        with self.assertRaises(hgvs.exceptions.HGVSInvalidIntervalError):
            (s, e) = ivm.map_tgt_to_ref(1, 200)
        with self.assertRaises(hgvs.exceptions.HGVSInvalidIntervalError):
            (s, e) = ivm.map_tgt_to_ref(0, 7)

    #
    # internal methods
    #

    def _check_valid_intervalpair(self, s1, e1, s2, e2):
        iv1 = hgvs.intervalmapper.Interval(s1, e1)
        iv2 = hgvs.intervalmapper.Interval(s2, e2)
        intervalpair = hgvs.intervalmapper.IntervalPair(iv1, iv2)
        repr_str = 'IntervalPair(ref=Interval(start_i={},end_i={}),tgt=Interval(start_i={},end_i={}))'
        self.assertEqual(intervalpair.__repr__(), repr_str.format(s1, e1, s2, e2))

    @staticmethod
    def _build_mock_intervalmapper():
        iv1 = [hgvs.intervalmapper.Interval(1, 5),
               hgvs.intervalmapper.Interval(5, 6),
               hgvs.intervalmapper.Interval(6, 10)
        ]

        iv2 = [hgvs.intervalmapper.Interval(1, 5),
               hgvs.intervalmapper.Interval(5, 5),
               hgvs.intervalmapper.Interval(5, 9)
        ]

        ivp = [hgvs.intervalmapper.IntervalPair(iv1[i], iv2[i]) for i in xrange(len(iv1))]
        ivm = hgvs.intervalmapper.IntervalMapper(ivp)
        return ivm



if __name__ == '__main__':
    unittest.main()

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/invitae/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>
